#------------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject_cache.pyx
#
# Cython file defining the DbObject cache implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

# SQL statements used within the DbObject type cache
cdef str DBO_CACHE_SQL_GET_METADATA_FOR_NAME = """
        declare
            t_Instantiable              varchar2(3);
            t_SuperTypeOwner            varchar2(128);
            t_SuperTypeName             varchar2(128);
            t_SubTypeRefCursor          sys_refcursor;
            t_Pos                       pls_integer;
        begin
            :ret_val := dbms_pickler.get_type_shape(:full_name, :oid,
                :version, :tds, t_Instantiable, t_SuperTypeOwner,
                t_SuperTypeName, :attrs_rc, t_SubTypeRefCursor);
            :package_name := null;
            if substr(:full_name, length(:full_name) - 7) = '%ROWTYPE' then
                t_Pos := instr(:full_name, '.');
                :schema := substr(:full_name, 1, t_Pos - 1);
                :name := substr(:full_name, t_Pos + 1);
            else
                begin
                    select owner, type_name
                    into :schema, :name
                    from all_types
                    where type_oid = :oid;
                exception
                when no_data_found then
                    begin
                        select owner, package_name, type_name
                        into :schema, :package_name, :name
                        from all_plsql_types
                        where type_oid = :oid;
                    exception
                    when no_data_found then
                        null;
                    end;
                end;
            end if;
        end;"""

cdef str DBO_CACHE_SQL_GET_COLUMNS = """
        select
            column_name,
            data_type,
            data_type_owner,
            case
                when data_type in
                        ('CHAR', 'NCHAR', 'VARCHAR2', 'NVARCHAR2', 'RAW')
                    then data_length
                else 0
            end,
            nvl(data_precision, 0),
            nvl(data_scale, 0)
        from all_tab_cols
        where owner = :owner
          and table_name = substr(:name, 1, length(:name) - 8)
          and hidden_column != 'YES'
        order by column_id"""

cdef str DBO_CACHE_SQL_GET_ELEM_TYPE_WITH_PACKAGE = """
        select elem_type_name
        from all_plsql_coll_types
        where owner = :owner
          and package_name = :package_name
          and type_name = :name"""

cdef str DBO_CACHE_SQL_GET_ELEM_TYPE_NO_PACKAGE = """
        select elem_type_name
        from all_coll_types
        where owner = :owner
          and type_name = :name"""

cdef str DBO_CACHE_SQL_GET_ELEM_OBJTYPE_WITH_PACKAGE = """
        select
            elem_type_owner,
            elem_type_package,
            elem_type_name
        from all_plsql_coll_types
        where owner = :owner
          and package_name = :package_name
          and type_name = :name"""

cdef str DBO_CACHE_SQL_GET_ELEM_OBJTYPE_NO_PACKAGE = """
        select
            elem_type_owner,
            elem_type_name
        from all_coll_types
        where owner = :owner
          and type_name = :name"""

cdef class ThinDbObjectTypeSuperCache:
    cdef:
        dict caches
        object lock
        int cache_num

    def __init__(self):
        self.caches = {}
        self.cache_num = 0
        self.lock = threading.Lock()


cdef class BaseThinDbObjectTypeCache:

    cdef:
        object meta_cursor, columns_cursor, attrs_ref_cursor_var, version_var
        object return_value_var, full_name_var, oid_var, tds_var
        object schema_var, package_name_var, name_var
        BaseThinConnImpl conn_impl
        dict types_by_oid
        dict types_by_name
        list partial_types

    cdef int _clear_cursors(self) except -1:
        """
        Clears the cursors used for searching metadata. This is needed when
        returning a connection to the pool since user-level objects are
        retained.
        """
        if self.meta_cursor is not None:
            self.meta_cursor.close()
            self.meta_cursor = None
            self.return_value_var = None
            self.full_name_var = None
            self.oid_var = None
            self.tds_var = None
            self.attrs_ref_cursor_var = None
            self.version_var = None
            self.schema_var = None
            self.package_name_var = None
            self.name_var = None
        if self.columns_cursor is not None:
            self.columns_cursor.close()
            self.columns_cursor = None

    cdef str _get_full_name(self, ThinDbObjectTypeImpl typ_impl):
        """
        Gets the full name of the type which is used for searching the database
        for the type information.
        """
        cdef str name, suffix = "%ROWTYPE"
        if typ_impl.name.endswith(suffix):
            name = typ_impl.name[:-len(suffix)]
        else:
            name = typ_impl.name
            suffix = ""
        if typ_impl.package_name is None:
            return f'"{typ_impl.schema}"."{name}"{suffix}'
        return f'"{typ_impl.schema}".' + \
               f'"{typ_impl.package_name}".' + \
               f'"{name}"{suffix}'

    cdef int _initialize(self, BaseThinConnImpl conn_impl) except -1:
        self.types_by_oid = {}
        self.types_by_name = {}
        self.partial_types = []
        self.conn_impl = conn_impl

    cdef int _init_columns_cursor(self, object conn) except -1:
        """
        Initializes the cursor that fetches the columns for a table or view.
        The input values come from the meta cursor that has been initialized
        and executed earlier.
        """
        cursor = conn.cursor()
        cursor.setinputsizes(owner=self.schema_var, name=self.name_var)
        cursor.prepare(DBO_CACHE_SQL_GET_COLUMNS)
        self.columns_cursor = cursor

    cdef int _init_meta_cursor(self, object conn) except -1:
        """
        Initializes the cursor that fetches the type metadata.
        """
        cursor = conn.cursor()
        self.return_value_var = cursor.var(DB_TYPE_BINARY_INTEGER)
        self.tds_var = cursor.var(bytes)
        self.full_name_var = cursor.var(str)
        self.schema_var = cursor.var(str)
        self.package_name_var = cursor.var(str)
        self.name_var = cursor.var(str)
        self.oid_var = cursor.var(bytes)
        self.version_var = cursor.var(DB_TYPE_BINARY_INTEGER)
        self.attrs_ref_cursor_var = cursor.var(DB_TYPE_CURSOR)
        cursor.setinputsizes(ret_val=self.return_value_var,
                             tds=self.tds_var,
                             full_name=self.full_name_var,
                             oid=self.oid_var,
                             schema=self.schema_var,
                             package_name=self.package_name_var,
                             name=self.name_var,
                             version=self.version_var,
                             attrs_rc=self.attrs_ref_cursor_var)
        cursor.prepare(DBO_CACHE_SQL_GET_METADATA_FOR_NAME)
        self.meta_cursor = cursor

    cdef object _parse_tds(self, ThinDbObjectTypeImpl typ_impl, bytes tds):
        """
        Parses the TDS for the type. This is only needed for collection types,
        so if the TDS is determined to be for an object type, the remaining
        information is skipped.
        """
        cdef:
            ThinDbObjectAttrImpl attr_impl
            int preferred_num_type
            uint16_t num_attrs, i
            uint8_t attr_type
            TDSBuffer buf
            uint32_t pos

        # parse initial TDS bytes
        buf = TDSBuffer.__new__(TDSBuffer)
        buf._populate_from_bytes(tds)
        buf.skip_raw_bytes(4)               # end offset
        buf.skip_raw_bytes(2)               # version op code and version
        buf.skip_raw_bytes(2)               # unknown
        buf.read_uint16(&num_attrs)         # number of attributes
        buf.skip_raw_bytes(1)               # TDS attributes?
        buf.skip_raw_bytes(1)               # start ADT op code
        buf.skip_raw_bytes(2)               # ADT number (always zero)
        buf.skip_raw_bytes(4)               # offset to index table

        # check to see if type refers to a collection (only one attribute is
        # present in that case)
        if num_attrs == 1:
            pos = buf._pos
            buf.read_ub1(&attr_type)
            if attr_type == TNS_OBJ_TDS_TYPE_COLL:
                typ_impl.is_collection = True
            else:
                buf.skip_to(pos)

        # handle collections
        if typ_impl.is_collection:
            buf.read_uint32(&pos)
            buf.read_uint32(&typ_impl.max_num_elements)
            buf.read_ub1(&typ_impl.collection_type)
            if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                typ_impl.collection_flags = TNS_OBJ_HAS_INDEXES
            buf.skip_to(pos)
            typ_impl.element_dbtype = self._parse_tds_attr(
                buf, &typ_impl.element_precision, &typ_impl.element_scale,
                &typ_impl.element_max_size,
                &typ_impl._element_preferred_num_type
            )
            if typ_impl.element_dbtype is DB_TYPE_CLOB:
                return self._get_element_type_clob(typ_impl)
            elif typ_impl.element_dbtype is DB_TYPE_OBJECT:
                return self._get_element_type_obj(typ_impl)

        # handle objects with attributes
        else:
            for i, attr_impl in enumerate(typ_impl.attrs):
                self._parse_tds_attr(buf, &attr_impl.precision,
                                     &attr_impl.scale, &attr_impl.max_size,
                                     &attr_impl._preferred_num_type)

    cdef DbType _parse_tds_attr(self, TDSBuffer buf, int8_t* precision,
                                int8_t* scale, uint32_t *max_size,
                                int* preferred_num_type):
        """
        Parses a TDS attribute from the buffer.
        """
        cdef:
            uint8_t attr_type, ora_type_num = 0, csfrm = 0
            int8_t temp_precision, temp_scale
            int temp_preferred_num_type
            uint32_t temp_max_size
            uint16_t temp16

        # skip until a type code that is of interest
        while True:
            buf.read_ub1(&attr_type)
            if attr_type == TNS_OBJ_TDS_TYPE_EMBED_ADT_INFO:
                buf.skip_raw_bytes(1)       # flags
            elif attr_type != TNS_OBJ_TDS_TYPE_SUBTYPE_MARKER:
                break

        # process the type code
        if attr_type == TNS_OBJ_TDS_TYPE_NUMBER:
            ora_type_num = TNS_DATA_TYPE_NUMBER
            buf.read_sb1(precision)
            buf.read_sb1(scale)
            preferred_num_type[0] = \
                    get_preferred_num_type(precision[0], scale[0])
        elif attr_type == TNS_OBJ_TDS_TYPE_FLOAT:
            ora_type_num = TNS_DATA_TYPE_NUMBER
            buf.skip_raw_bytes(1)           # precision
        elif attr_type in (TNS_OBJ_TDS_TYPE_VARCHAR, TNS_OBJ_TDS_TYPE_CHAR):
            buf.read_uint16(&temp16)        # maximum length
            max_size[0] = temp16
            buf.read_ub1(&csfrm)
            csfrm = csfrm & 0x7f
            buf.skip_raw_bytes(2)           # character set
            if attr_type == TNS_OBJ_TDS_TYPE_VARCHAR:
                ora_type_num = TNS_DATA_TYPE_VARCHAR
            else:
                ora_type_num = TNS_DATA_TYPE_CHAR
        elif attr_type == TNS_OBJ_TDS_TYPE_RAW:
            buf.read_uint16(&temp16)        # maximum length
            max_size[0] = temp16
            ora_type_num = TNS_DATA_TYPE_RAW
        elif attr_type == TNS_OBJ_TDS_TYPE_BINARY_FLOAT:
            ora_type_num = TNS_DATA_TYPE_BINARY_FLOAT
        elif attr_type == TNS_OBJ_TDS_TYPE_BINARY_DOUBLE:
            ora_type_num = TNS_DATA_TYPE_BINARY_DOUBLE
        elif attr_type == TNS_OBJ_TDS_TYPE_DATE:
            ora_type_num = TNS_DATA_TYPE_DATE
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP:
            buf.skip_raw_bytes(1)           # precision
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP_LTZ:
            buf.skip_raw_bytes(1)           # precision
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP_LTZ
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP_TZ:
            buf.skip_raw_bytes(1)           # precision
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP_TZ
        elif attr_type == TNS_OBJ_TDS_TYPE_BOOLEAN:
            ora_type_num = TNS_DATA_TYPE_BOOLEAN
        elif attr_type == TNS_OBJ_TDS_TYPE_CLOB:
            ora_type_num = TNS_DATA_TYPE_CLOB
            csfrm = CS_FORM_IMPLICIT
        elif attr_type == TNS_OBJ_TDS_TYPE_BLOB:
            ora_type_num = TNS_DATA_TYPE_BLOB
        elif attr_type == TNS_OBJ_TDS_TYPE_OBJ:
            ora_type_num = TNS_DATA_TYPE_INT_NAMED
            buf.skip_raw_bytes(5)           # offset and code
        elif attr_type == TNS_OBJ_TDS_TYPE_START_EMBED_ADT:
            ora_type_num = TNS_DATA_TYPE_INT_NAMED
            while self._parse_tds_attr(buf, &temp_precision, &temp_scale,
                                       &temp_max_size,
                                       &temp_preferred_num_type):
                pass
        elif attr_type == TNS_OBJ_TDS_TYPE_END_EMBED_ADT:
            return None
        else:
            errors._raise_err(errors.ERR_TDS_TYPE_NOT_SUPPORTED, num=attr_type)
        return DbType._from_ora_type_and_csfrm(ora_type_num, csfrm)

    cdef int _create_attr(self, ThinDbObjectTypeImpl typ_impl, str name,
                          str type_name, str type_owner,
                          str type_package_name=None, bytes oid=None,
                          int8_t precision=0, int8_t scale=0,
                          uint32_t max_size=0) except -1:
        """
        Creates an attribute from the supplied information and adds it to the
        list of attributes for the type.
        """
        cdef:
            ThinDbObjectTypeImpl attr_typ_impl
            ThinDbObjectAttrImpl attr_impl
        attr_impl = ThinDbObjectAttrImpl.__new__(ThinDbObjectAttrImpl)
        attr_impl.name = name
        if type_owner is not None:
            attr_typ_impl = self.get_type_for_info(oid, type_owner,
                                                   type_package_name,
                                                   type_name)
            if attr_typ_impl.is_xml_type:
                attr_impl.dbtype = DB_TYPE_XMLTYPE
            else:
                attr_impl.dbtype = DB_TYPE_OBJECT
                attr_impl.objtype = attr_typ_impl
        else:
            attr_impl.dbtype = DbType._from_ora_name(type_name)
            attr_impl.max_size = max_size
            if precision != 0 or scale != 0:
                attr_impl.precision = precision
                attr_impl.scale = scale
                attr_impl._preferred_num_type = \
                        get_preferred_num_type(precision, scale)
        typ_impl.attrs.append(attr_impl)
        typ_impl.attrs_by_name[name] = attr_impl

    cdef object _populate_type_info(self, str name, object attrs,
                                    ThinDbObjectTypeImpl typ_impl):
        """
        Populate the type information given the name of the type.
        """
        cdef:
            ssize_t start_pos, end_pos, name_length
            ThinDbObjectAttrImpl attr_impl
            str data_type
        typ_impl.version = self.version_var.getvalue()
        if typ_impl.oid is None:
            typ_impl.oid = self.oid_var.getvalue()
            self.types_by_oid[typ_impl.oid] = typ_impl
        if typ_impl.schema is None:
            typ_impl.schema = self.schema_var.getvalue()
            typ_impl.package_name = self.package_name_var.getvalue()
            typ_impl.name = self.name_var.getvalue()
            typ_impl.is_xml_type = \
                    (typ_impl.schema == "SYS" and typ_impl.name == "XMLTYPE")
        typ_impl.attrs = []
        typ_impl.attrs_by_name = {}
        if typ_impl.is_row_type:
            for name, data_type, data_type_owner, max_size, precision, \
                    scale in attrs:
                if data_type_owner is None:
                    start_pos = data_type.find("(")
                    if start_pos > 0:
                        end_pos = data_type.find(")")
                        if end_pos > start_pos:
                            data_type = data_type[:start_pos] + \
                                    data_type[end_pos + 1:]
                self._create_attr(typ_impl, name, data_type, data_type_owner,
                                  None, None, precision, scale, max_size)
        else:
            for cursor_version, attr_name, attr_num, attr_type_name, \
                    attr_type_owner, attr_type_package, attr_type_oid, \
                    attr_instantiable, attr_super_type_owner, \
                    attr_super_type_name in attrs:
                if attr_name is None:
                    continue
                self._create_attr(typ_impl, attr_name, attr_type_name,
                                  attr_type_owner, attr_type_package,
                                  attr_type_oid)
            return self._parse_tds(typ_impl, self.tds_var.getvalue())

    cdef ThinDbObjectTypeImpl get_type_for_info(self, bytes oid, str schema,
                                                str package_name, str name):
        """
        Returns a type for the specified fetch info, if one has already been
        cached. If not, a new type object is created and cached. It is also
        added to the partial_types list which will be fully populated once the
        current execute has completed.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            str full_name
        if package_name is not None:
            full_name = f"{schema}.{package_name}.{name}"
        else:
            full_name = f"{schema}.{name}"
        if oid is not None:
            typ_impl = self.types_by_oid.get(oid)
        else:
            typ_impl = self.types_by_name.get(full_name)
        if typ_impl is None:
            typ_impl = ThinDbObjectTypeImpl.__new__(ThinDbObjectTypeImpl)
            typ_impl._conn_impl = self.conn_impl
            typ_impl.oid = oid
            typ_impl.schema = schema
            typ_impl.package_name = package_name
            typ_impl.name = name
            typ_impl.is_xml_type = (schema == "SYS" and name == "XMLTYPE")
            if oid is not None:
                self.types_by_oid[oid] = typ_impl
            self.types_by_name[full_name] = typ_impl
            self.partial_types.append(typ_impl)
        return typ_impl


cdef class ThinDbObjectTypeCache(BaseThinDbObjectTypeCache):

    def _get_element_type_clob(self, ThinDbObjectTypeImpl typ_impl):
        """
        Determine if the element type refers to an NCLOB or CLOB value. This
        must be fetched from the data dictionary since it is not included in
        the TDS.
        """
        cursor = self.meta_cursor.connection.cursor()
        if typ_impl.package_name is not None:
            cursor.execute(DBO_CACHE_SQL_GET_ELEM_TYPE_WITH_PACKAGE,
                    owner=typ_impl.schema,
                    package_name=typ_impl.package_name,
                    name=typ_impl.name)
        else:
            cursor.execute(DBO_CACHE_SQL_GET_ELEM_TYPE_NO_PACKAGE,
                    owner=typ_impl.schema,
                    name=typ_impl.name)
        type_name, = cursor.fetchone()
        if type_name == "NCLOB":
            typ_impl.element_dbtype = DB_TYPE_NCLOB

    def _get_element_type_obj(self, ThinDbObjectTypeImpl typ_impl):
        """
        Determine the element type's object type. This is needed when
        processing collections with object as the element type since this
        information is not available in the TDS.
        """
        cdef:
            str schema, name, package_name = None
            object cursor
        cursor = self.meta_cursor.connection.cursor()
        if typ_impl.package_name is not None:
            cursor.execute(DBO_CACHE_SQL_GET_ELEM_OBJTYPE_WITH_PACKAGE,
                    owner=typ_impl.schema,
                    package_name=typ_impl.package_name,
                    name=typ_impl.name)
            schema, package_name, name = cursor.fetchone()
        else:
            cursor.execute(DBO_CACHE_SQL_GET_ELEM_OBJTYPE_NO_PACKAGE,
                    owner=typ_impl.schema,
                    name=typ_impl.name)
            schema, name = cursor.fetchone()
        typ_impl.element_objtype = self.get_type_for_info(None, schema,
                                                          package_name, name)

    cdef list _lookup_type(self, object conn, str name,
                           ThinDbObjectTypeImpl typ_impl):
        """
        Lookup the type given its name and return the list of attributes for
        further processing. The metadata cursor execution will populate the
        variables.
        """
        if self.meta_cursor is None:
            self._init_meta_cursor(conn)
        self.full_name_var.setvalue(0, name)
        self.meta_cursor.execute(None)
        if self.return_value_var.getvalue() != 0:
            errors._raise_err(errors.ERR_INVALID_OBJECT_TYPE_NAME, name=name)
        if name.endswith("%ROWTYPE"):
            typ_impl.is_row_type = True
            if self.columns_cursor is None:
                self._init_columns_cursor(conn)
            self.columns_cursor.execute(None)
            return self.columns_cursor.fetchall()
        else:
            attrs_rc = self.attrs_ref_cursor_var.getvalue()
            return attrs_rc.fetchall()

    cdef ThinDbObjectTypeImpl get_type(self, object conn, str name):
        """
        Returns the database object type given its name. The cache is first
        searched and if it is not found, the database is searched and the
        result stored in the cache.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            bint is_rowtype
        typ_impl = self.types_by_name.get(name)
        if typ_impl is None:
            typ_impl = ThinDbObjectTypeImpl.__new__(ThinDbObjectTypeImpl)
            typ_impl._conn_impl = self.conn_impl
            attrs = self._lookup_type(conn, name, typ_impl)
            self._populate_type_info(name, attrs, typ_impl)
            self.types_by_oid[typ_impl.oid] = typ_impl
            self.types_by_name[name] = typ_impl
            self.populate_partial_types(conn)
        return typ_impl

    def populate_partial_types(self, object conn):
        """
        Populate any partial types that were discovered earlier. Since
        populating an object type might result in additional object types being
        discovered, object types are popped from the partial types list until
        the list is empty.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            str full_name
            list attrs
        while self.partial_types:
            typ_impl = self.partial_types.pop()
            full_name = self._get_full_name(typ_impl)
            attrs = self._lookup_type(conn, full_name, typ_impl)
            self._populate_type_info(full_name, attrs, typ_impl)


cdef class AsyncThinDbObjectTypeCache(BaseThinDbObjectTypeCache):

    async def _get_element_type_clob(self, ThinDbObjectTypeImpl typ_impl):
        """
        Determine if the element type refers to an NCLOB or CLOB value. This
        must be fetched from the data dictionary since it is not included in
        the TDS.
        """
        cursor = self.meta_cursor.connection.cursor()
        if typ_impl.package_name is not None:
            await cursor.execute(DBO_CACHE_SQL_GET_ELEM_TYPE_WITH_PACKAGE,
                    owner=typ_impl.schema,
                    package_name=typ_impl.package_name,
                    name=typ_impl.name)
        else:
            await cursor.execute(DBO_CACHE_SQL_GET_ELEM_TYPE_NO_PACKAGE,
                    owner=typ_impl.schema,
                    name=typ_impl.name)
        type_name, = await cursor.fetchone()
        if type_name == "NCLOB":
            typ_impl.element_dbtype = DB_TYPE_NCLOB

    async def _get_element_type_obj(self, ThinDbObjectTypeImpl typ_impl):
        """
        Determine the element type's object type. This is needed when
        processing collections with object as the element type since this
        information is not available in the TDS.
        """
        cdef:
            str schema, name, package_name = None
            object cursor
        cursor = self.meta_cursor.connection.cursor()
        if typ_impl.package_name is not None:
            await cursor.execute(DBO_CACHE_SQL_GET_ELEM_OBJTYPE_WITH_PACKAGE,
                    owner=typ_impl.schema,
                    package_name=typ_impl.package_name,
                    name=typ_impl.name)
            schema, package_name, name = await cursor.fetchone()
        else:
            await cursor.execute(DBO_CACHE_SQL_GET_ELEM_OBJTYPE_NO_PACKAGE,
                    owner=typ_impl.schema,
                    name=typ_impl.name)
            schema, name = await cursor.fetchone()
        typ_impl.element_objtype = self.get_type_for_info(None, schema,
                                                          package_name, name)

    async def _lookup_type(self, object conn, str name,
                           ThinDbObjectTypeImpl typ_impl):
        """
        Lookup the type given its name and return the list of attributes for
        further processing. The metadata cursor execution will populate the
        variables.
        """
        if self.meta_cursor is None:
            self._init_meta_cursor(conn)
        self.full_name_var.setvalue(0, name)
        await self.meta_cursor.execute(None)
        if self.return_value_var.getvalue() != 0:
            errors._raise_err(errors.ERR_INVALID_OBJECT_TYPE_NAME, name=name)
        if name.endswith("%ROWTYPE"):
            typ_impl.is_row_type = True
            if self.columns_cursor is None:
                self._init_columns_cursor(conn)
            await self.columns_cursor.execute(None)
            return await self.columns_cursor.fetchall()
        else:
            attrs_rc = self.attrs_ref_cursor_var.getvalue()
            return await attrs_rc.fetchall()

    async def get_type(self, object conn, str name):
        """
        Returns the database object type given its name. The cache is first
        searched and if it is not found, the database is searched and the
        result stored in the cache.
        """
        cdef ThinDbObjectTypeImpl typ_impl
        typ_impl = self.types_by_name.get(name)
        if typ_impl is None:
            typ_impl = ThinDbObjectTypeImpl.__new__(ThinDbObjectTypeImpl)
            typ_impl._conn_impl = self.conn_impl
            attrs = await self._lookup_type(conn, name, typ_impl)
            coroutine = self._populate_type_info(name, attrs, typ_impl)
            if coroutine is not None:
                await coroutine
            self.types_by_oid[typ_impl.oid] = typ_impl
            self.types_by_name[name] = typ_impl
            await self.populate_partial_types(conn)
        return typ_impl

    async def populate_partial_types(self, object conn):
        """
        Populate any partial types that were discovered earlier. Since
        populating an object type might result in additional object types being
        discovered, object types are popped from the partial types list until
        the list is empty.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            str full_name
            list attrs
        while self.partial_types:
            typ_impl = self.partial_types.pop()
            full_name = self._get_full_name(typ_impl)
            attrs = await self._lookup_type(conn, full_name, typ_impl)
            coroutine = self._populate_type_info(full_name, attrs, typ_impl)
            if coroutine is not None:
                await coroutine


# global cache of database object types
# since the database object types require a reference to the connection (in
# order to be able to manage LOBs), storing the cache on the connection would
# involve creating a circular reference
cdef ThinDbObjectTypeSuperCache DB_OBJECT_TYPE_SUPER_CACHE = \
        ThinDbObjectTypeSuperCache()


cdef int create_new_dbobject_type_cache(BaseThinConnImpl conn_impl) except -1:
    """
    Creates a new database object type cache and returns its identifier.
    """
    cdef:
        BaseThinDbObjectTypeCache cache
        bint is_async
        int cache_num
    with DB_OBJECT_TYPE_SUPER_CACHE.lock:
        DB_OBJECT_TYPE_SUPER_CACHE.cache_num += 1
        cache_num = DB_OBJECT_TYPE_SUPER_CACHE.cache_num
    is_async = conn_impl._protocol._transport._is_async
    cls = AsyncThinDbObjectTypeCache if is_async else ThinDbObjectTypeCache
    cache = cls.__new__(cls)
    cache._initialize(conn_impl)
    DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num] = cache
    return cache_num


cdef BaseThinDbObjectTypeCache get_dbobject_type_cache(int cache_num):
    """
    Returns the database object type cache given its identifier.
    """
    return DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num]


cdef int remove_dbobject_type_cache(int cache_num) except -1:
    """
    Removes the sub cache given its identifier.
    """
    del DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num]



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Loads the test environment found in the base directory and then extends it
# with some methods used to determine which extended tests to run.
# -----------------------------------------------------------------------------

import configparser
import os
import secrets
import string

dir_name = os.path.dirname(os.path.dirname(__file__))
file_name = os.path.join(dir_name, os.path.basename(__file__))
exec(open(file_name).read(), globals(), locals())

DATABASES_SECTION_NAME = "Databases"


class ExtendedConfig:

    def __init__(self):
        default_file_name = os.path.join(
            os.path.dirname(__file__), "config.ini"
        )
        file_name = os.environ.get(
            "PYO_TEST_EXT_CONFIG_FILE", default_file_name
        )
        self.parser = configparser.ConfigParser()
        self.parser.read(file_name)
        self.section_name = "DEFAULT"
        connect_string_to_use = get_connect_string().upper()  # noqa: F821

        if self.parser.has_section(DATABASES_SECTION_NAME):
            for section_name, connect_string in self.parser.items(
                DATABASES_SECTION_NAME
            ):
                if connect_string.upper() == connect_string_to_use:
                    self.section_name = section_name
                    break


_extended_config = ExtendedConfig()


def get_extended_config_bool(name, fallback=False):
    return _extended_config.parser.getboolean(
        _extended_config.section_name, name, fallback=fallback
    )


def get_extended_config_str(name, fallback=None):
    return _extended_config.parser.get(
        _extended_config.section_name, name, fallback=fallback
    )


def get_random_string(length=10):
    return "".join(secrets.choice(string.ascii_letters) for i in range(length))



# -----------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
4700 - Module for testing pool parameters.
"""

import ssl

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def __test_writable_parameter(self, name, value):
        """
        Tests that a writable parameter can be written to and the modified
        value read back successfully.
        """
        params = oracledb.PoolParams()
        orig_value = getattr(params, name)
        copied_params = params.copy()
        args = {}
        args[name] = value
        params.set(**args)
        self.assertEqual(getattr(params, name), value)
        self.assertEqual(getattr(copied_params, name), orig_value)

    def test_4700(self):
        "4700 - test writable parameters"
        self.__test_writable_parameter("min", 8)
        self.__test_writable_parameter("max", 12)
        self.__test_writable_parameter("increment", 2)
        self.__test_writable_parameter("connectiontype", oracledb.Connection)
        self.__test_writable_parameter("getmode", oracledb.POOL_GETMODE_NOWAIT)
        self.__test_writable_parameter("homogeneous", False)
        self.__test_writable_parameter("timeout", 25)
        self.__test_writable_parameter("wait_timeout", 45)
        self.__test_writable_parameter("max_lifetime_session", 65)
        self.__test_writable_parameter("session_callback", lambda c: None)
        self.__test_writable_parameter("max_sessions_per_shard", 5)
        self.__test_writable_parameter("soda_metadata_cache", True)
        self.__test_writable_parameter("ping_interval", 20)
        self.__test_writable_parameter("ping_timeout", 3000)

    def test_4701(self):
        "4701 - test PoolParams repr()"
        values = [
            ("min", 3),
            ("max", 10),
            ("increment", 4),
            ("connectiontype", oracledb.Connection),
            ("getmode", oracledb.POOL_GETMODE_WAIT),
            ("homogeneous", True),
            ("timeout", 60),
            ("wait_timeout", 20),
            ("max_lifetime_session", 80),
            ("session_callback", lambda c: None),
            ("max_sessions_per_shard", 4),
            ("soda_metadata_cache", False),
            ("ping_interval", 50),
            ("ping_timeout", 2500),
            ("user", test_env.get_main_user()),
            ("proxy_user", test_env.get_proxy_user()),
            ("host", "my_host1"),
            ("port", 1522),
            ("protocol", "tcp"),
            ("https_proxy", "proxy_4701"),
            ("https_proxy_port", 4701),
            ("service_name", "my_service_name1"),
            ("sid", "my_sid1"),
            ("server_type", "dedicated"),
            ("cclass", "cclass_1"),
            ("purity", oracledb.PURITY_SELF),
            ("expire_time", 60),
            ("retry_count", 6),
            ("retry_delay", 10),
            ("tcp_connect_timeout", 40.0),
            ("ssl_server_dn_match", False),
            ("ssl_server_cert_dn", "CN=unknown4701a"),
            ("wallet_location", "/tmp/wallet_loc1a"),
            ("events", True),
            ("externalauth", True),
            ("mode", oracledb.AUTH_MODE_SYSDBA),
            ("disable_oob", True),
            ("stmtcachesize", 25),
            ("edition", "edition_4701"),
            ("tag", "tag4701"),
            ("matchanytag", True),
            ("config_dir", "config_dir_4701"),
            ("appcontext", [("a", "b", "c")]),
            ("shardingkey", [1, 2, 3]),
            ("supershardingkey", [4]),
            ("debug_jdwp", "host=host;port=1523"),
            ("connection_id_prefix", "prefix4701"),
            ("ssl_context", None),
            ("sdu", 16384),
            ("pool_boundary", "transaction"),
            ("use_tcp_fast_open", True),
            ("ssl_version", ssl.TLSVersion.TLSv1_2),
        ]
        params = oracledb.PoolParams(**dict(values))
        parts = [f"{name}={value!r}" for name, value in values]
        expected_value = f"PoolParams({', '.join(parts)})"
        self.assertEqual(repr(params), expected_value)
        self.assertIs(params.getmode, oracledb.PoolGetMode.WAIT)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# statement_cache.pyx
#
# Cython file defining the StatementCache class used to manage cached
# statements (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class StatementCache:

    cdef:
        object _cached_statements
        object _lock
        uint32_t _max_size
        uint32_t _max_cursors
        array.array _cursors_to_close
        ssize_t _num_cursors_to_close
        set _open_cursors

    cdef int _add_cursor_to_close(self, Statement stmt) except -1:
        """
        Add the statement's cursor to the list of cursors that need to be
        closed.
        """
        if stmt._cursor_id != 0:
            self._cursors_to_close[self._num_cursors_to_close] = \
                    stmt._cursor_id
            self._num_cursors_to_close += 1
        self._open_cursors.remove(stmt)

    cdef int _adjust_cache(self) except -1:
        """
        Adjust the cache so that no more than the maximum number of statements
        are cached.
        """
        cdef Statement stmt
        while len(self._cached_statements) > self._max_size:
            stmt = <Statement> self._cached_statements.popitem(last=False)[1]
            if stmt._in_use:
                stmt._return_to_cache = False
            else:
                self._add_cursor_to_close(stmt)

    cdef int clear_cursor(self, Statement stmt) except -1:
        """
        Clears the particular cursor but retains it in the list of open
        cursors. This is for cases where the cursor id needs to be cleared but
        the cursor itself is still going to be used.
        """
        self._cursors_to_close[self._num_cursors_to_close] = stmt._cursor_id
        self._num_cursors_to_close += 1
        stmt._cursor_id = 0
        stmt._executed = False

    cdef int clear_open_cursors(self) except -1:
        """
        Clears the list of open cursors and removes the list of cursors that
        need to be closed. This is required when a DRCP session change has
        taken place as the cursor ID values are invalidated.
        """
        cdef:
            set new_open_cursors = set()
            Statement stmt
        with self._lock:
            for stmt in self._open_cursors:
                stmt.clear_all_state()
                if stmt._in_use or stmt._return_to_cache:
                    new_open_cursors.add(stmt)
            self._open_cursors = new_open_cursors
            self._num_cursors_to_close = 0

    cdef Statement get_statement(self, str sql, bint cache_statement,
                                 bint force_new):
        """
        Get a statement from the statement cache, or prepare a new statement
        for use. If a statement is already in use or the statement is not
        supposed to be cached, a copy will be made (and not returned to the
        cache).
        """
        cdef Statement stmt = None
        with self._lock:
            if sql is not None:
                stmt = self._cached_statements.get(sql)
            if stmt is None:
                stmt = Statement.__new__(Statement)
                if sql is not None:
                    stmt._prepare(sql)
                if cache_statement and not stmt._is_ddl and self._max_size > 0:
                    stmt._return_to_cache = True
                    self._cached_statements[sql] = stmt
                    self._adjust_cache()
                self._open_cursors.add(stmt)
            elif force_new or stmt._in_use:
                if not cache_statement:
                    self._add_cursor_to_close(stmt)
                    del self._cached_statements[sql]
                stmt = stmt.copy()
                self._open_cursors.add(stmt)
            elif not cache_statement:
                del self._cached_statements[sql]
                stmt._return_to_cache = False
            else:
                self._cached_statements.move_to_end(sql)
        stmt._in_use = True
        return stmt

    cdef int initialize(self, uint32_t max_size,
                         uint32_t max_cursors) except -1:
        """
        Initialize the statement cache.
        """
        if max_cursors == 0:
            self._max_size = 0
            self._max_cursors = 1
        else:
            self._max_size = max_size
            self._max_cursors = max_cursors
        self._cached_statements = collections.OrderedDict()
        self._lock = threading.Lock()
        self._open_cursors = set()
        self._cursors_to_close = array.array('I')
        array.resize(self._cursors_to_close, self._max_cursors)
        self._num_cursors_to_close = 0

    cdef int resize(self, uint32_t new_size) except -1:
        """
        Resizes the statement cache to the new value.
        """
        with self._lock:
            self._max_size = new_size
            self._adjust_cache()

    cdef int return_statement(self, Statement stmt) except -1:
        """
        Return the statement to the statement cache, if applicable. If the
        statement must not be returned to the statement cache, add the cursor
        id to the list of cursor ids to close on the next round trip to the
        database. Clear all bind variables and fetch variables in order to
        ensure that unnecessary references are not retained.
        """
        cdef:
            ThinVarImpl var_impl
            BindInfo bind_info
        if stmt._bind_info_list is not None:
            for bind_info in stmt._bind_info_list:
                bind_info._bind_var_impl = None
        if stmt._fetch_var_impls is not None:
            for var_impl in stmt._fetch_var_impls:
                var_impl._values = [None] * var_impl.num_elements
        with self._lock:
            if stmt._return_to_cache:
                stmt._in_use = False
            else:
                self._add_cursor_to_close(stmt)

    cdef int write_cursors_to_close(self, WriteBuffer buf) except -1:
        """
        Write the list of cursors to close to the buffer.
        """
        cdef:
            unsigned int *cursor_ids
            ssize_t i
        with self._lock:
            buf.write_ub4(self._num_cursors_to_close)
            cursor_ids = self._cursors_to_close.data.as_uints
            for i in range(self._num_cursors_to_close):
                buf.write_ub4(cursor_ids[i])
            self._num_cursors_to_close = 0



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6900 - Module for testing OSON encoding and decoding.
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (21, 0), "unsupported client"
)
@unittest.skipUnless(
    test_env.get_server_version() >= (21, 0), "unsupported server"
)
class TestCase(test_env.BaseTestCase):
    def test_6900(self):
        "6900 - test OSON metadata"
        self.cursor.execute("select * from TestOsonCols")
        int_col_metadata, oson_col_metadata = self.cursor.description
        self.assertFalse(int_col_metadata.is_oson)
        self.assertTrue(oson_col_metadata.is_oson)

    def test_6901(self):
        "6901 - test simple query of OSON encoded bytes"
        self.cursor.execute("delete from TestOsonCols")
        self.cursor.execute(
            """
                insert into TestOsonCols (IntCol, OsonCol)
                values (1, '{"id": 6901, "value" : "string 6901"}')"""
        )
        self.conn.commit()
        self.cursor.execute("select OsonCol from TestOsonCols")
        (oson_val,) = self.cursor.fetchone()
        expected_val = dict(id=6901, value="string 6901")
        self.assertEqual(oson_val, expected_val)

    def test_6902(self):
        "6902 - test round trip of OSON encoded bytes"
        value = dict(id=6902, value="string 6902")
        self.cursor.execute("delete from TestOsonCols")
        encoded_oson = self.conn.encode_oson(value)
        self.cursor.execute(
            "insert into TestOsonCols values (1, :data)", [encoded_oson]
        )
        self.conn.commit()
        self.cursor.execute("select OsonCol from TestOsonCols")
        (oson_val,) = self.cursor.fetchone()
        self.assertEqual(oson_val, value)

    def test_6903(self):
        "6903 - test encoding and decoding a value"
        value = dict(id=6903, value="string 6903")
        out_value = self.conn.decode_oson(self.conn.encode_oson(value))
        self.assertEqual(out_value, value)

    def test_6904(self):
        "6904 - test decoding a non encoded value"
        value = b"{'not a previous encoded value': 3}"
        with self.assertRaisesFullCode("DPY-5004"):
            self.conn.decode_oson(value)

    def test_6905(self):
        "6905 - test inserting oson inside a lob"
        value = dict(id=6905, value="string 6905")
        self.cursor.execute("delete from TestOsonCols")
        encoded_oson = self.conn.encode_oson(value)
        lob = self.conn.createlob(oracledb.DB_TYPE_BLOB, encoded_oson)
        self.cursor.execute(
            "insert into TestOsonCols values (1, :data)", [lob]
        )
        self.conn.commit()
        self.cursor.execute("select OsonCol from TestOsonCols")
        (oson_val,) = self.cursor.fetchone()
        self.assertEqual(oson_val, value)

    def test_6906(self):
        "6906 - test inserting oson as json"
        self.cursor.execute("delete from TestOsonCols")
        value = dict(id=6906, value="string 6906")
        oson = self.conn.encode_oson(value)
        self.cursor.setinputsizes(oracledb.DB_TYPE_JSON)
        self.cursor.execute(
            "insert into TestOsonCols values (1, :data)", [oson]
        )
        self.conn.commit()
        self.cursor.execute("select OsonCol from TestOsonCols")
        (oson_val,) = self.cursor.fetchone()
        oson_val = self.conn.decode_oson(oson_val)
        self.assertEqual(oson_val, value)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3400 - Module for testing Simple Oracle Document Access (SODA) Collections
"""

import datetime
import decimal
import json
import unittest

import oracledb
import test_env


@unittest.skipIf(
    test_env.skip_soda_tests(), "unsupported client/server combination"
)
class TestCase(test_env.BaseTestCase):
    def __normalize_docs(self, docs):
        """
        Remove the embedded OID added in Oracle Database 23ai, if found, in
        order to ease comparison.
        """
        for doc in docs:
            if doc is not None and "_id" in doc:
                del doc["_id"]

    def __test_skip(self, coll, num_to_skip, expected_content):
        filter_spec = {"$orderby": [{"path": "name", "order": "desc"}]}
        doc = coll.find().filter(filter_spec).skip(num_to_skip).getOne()
        content = doc.getContent() if doc is not None else None
        self.__normalize_docs([content])
        self.assertEqual(content, expected_content)

    def test_3400(self):
        "3400 - test inserting invalid JSON value into SODA collection"
        invalid_json = "{testKey:testValue}"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("InvalidJSON")
        doc = soda_db.createDocument(invalid_json)
        with self.assertRaisesFullCode("ORA-40780", "ORA-02290"):
            coll.insertOne(doc)

    def test_3401(self):
        "3401 - test inserting documents into a SODA collection"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestInsertDocs")
        values_to_insert = [
            {"name": "George", "age": 47},
            {"name": "Susan", "age": 39},
            {"name": "John", "age": 50},
            {"name": "Jill", "age": 54},
        ]
        inserted_keys = []
        for value in values_to_insert:
            doc = coll.insertOneAndGet(value)
            inserted_keys.append(doc.key)
        self.conn.commit()
        self.assertEqual(coll.find().count(), len(values_to_insert))
        for key, value in zip(inserted_keys, values_to_insert):
            doc = coll.find().key(key).getOne().getContent()
            self.__normalize_docs([doc])
            self.assertEqual(doc, value)

    def test_3402(self):
        "3402 - test skipping documents in a SODA collection"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSkipDocs")
        values_to_insert = [
            {"name": "Anna", "age": 62},
            {"name": "Mark", "age": 37},
            {"name": "Martha", "age": 43},
            {"name": "Matthew", "age": 28},
        ]
        for value in values_to_insert:
            coll.insertOne(value)
        self.conn.commit()
        self.__test_skip(coll, 0, values_to_insert[3])
        self.__test_skip(coll, 1, values_to_insert[2])
        self.__test_skip(coll, 3, values_to_insert[0])
        self.__test_skip(coll, 4, None)
        self.__test_skip(coll, 125, None)

    def test_3403(self):
        "3403 - test replace documents in SODA collection"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestReplaceDoc")
        content = {"name": "John", "address": {"city": "Sydney"}}
        doc = coll.insertOneAndGet(content)
        new_content = {"name": "John", "address": {"city": "Melbourne"}}
        replaced = coll.find().key(doc.key).replaceOne(new_content)
        self.assertTrue(replaced)
        self.conn.commit()
        doc = coll.find().key(doc.key).getOne().getContent()
        self.__normalize_docs([doc])
        self.assertEqual(doc, new_content)

    def test_3404(self):
        "3404 - test search documents with different QBEs"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSearchDocContent")
        data = [
            {
                "name": "John",
                "age": 22,
                "birthday": "2000-12-15",
                "locations": [{"city": "Bangalore"}, {"city": "Other"}],
            },
            {
                "name": "Johnson",
                "age": 45,
                "birthday": "1978-02-03",
                "locations": [{"city": "Banaras"}, {"city": "Manhattan"}],
            },
            {
                "name": "William",
                "age": 32,
                "birthday": "1991-05-17",
                "locations": {"city": "New Banaras"},
            },
        ]
        coll.insertMany(data)
        self.conn.commit()

        # create index so $contains works
        index = {"name": "js_ix_3404"}
        coll.createIndex(index)

        filter_specs = [
            ({"name": {"$contains": "John"}}, 1),
            ({"age": {"$contains": "45"}}, 1),
            ({"name": {"$like": "J%n"}}, 2),
            ({"name": {"$regex": ".*[ho]n"}}, 2),
            ("""{"locations.city": {"$regex": "^Ban.*"}}""", 2),
            ({"birthday": {"$date": {"$gt": "2000-01-01"}}}, 1),
            ({"birthday": {"$date": "2000-12-15"}}, 1),
            ({"age": {"$gt": 18}}, 3),
            ({"age": {"$lt": 25}}, 1),
            (
                {
                    "$or": [
                        {"age": {"$gt": 50}},
                        {"locations[*].city": {"$like": "%Ban%"}},
                    ]
                },
                3,
            ),
            (
                {
                    "$and": [
                        {"age": {"$gt": 40}},
                        {"locations[0 to 1].city": {"$like": "%aras"}},
                    ]
                },
                1,
            ),
            ({"name": {"$hasSubstring": "John"}}, 2),
            ({"name": {"$instr": "John"}}, 2),
            ({"name": {"$startsWith": "John"}}, 2),
            ({"name": {"$upper": {"$startsWith": "JO"}}}, 2),
            ({"age": {"$not": {"$eq": 22}}}, 2),
            ({"age": {"$not": {"$lt": 30, "$gt": 10}}}, 2),
            ({"locations": {"$type": "array"}}, 2),
        ]
        for filter_spec, expected_count in filter_specs:
            self.assertEqual(
                coll.find().filter(filter_spec).count(),
                expected_count,
                filter_spec,
            )

    def test_3405(self):
        "3405 - test removing documents"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestRemoveDocs")
        data = [
            {"name": "John", "address": {"city": "Bangalore"}},
            {"name": "Johnson", "address": {"city": "Banaras"}},
            {"name": "Joseph", "address": {"city": "Mangalore"}},
            {"name": "Jibin", "address": {"city": "Secunderabad"}},
            {"name": "Andrew", "address": {"city": "Hyderabad"}},
            {"name": "Matthew", "address": {"city": "Mumbai"}},
        ]
        docs = [coll.insertOneAndGet(v) for v in data]
        self.assertEqual(coll.find().key(docs[3].key).remove(), 1)
        self.assertEqual(coll.find().count(), len(data) - 1)
        search_results = coll.find().filter({"name": {"$like": "Jibin"}})
        self.assertEqual(search_results.count(), 0)
        self.assertEqual(
            coll.find().filter({"name": {"$like": "John%"}}).remove(), 2
        )
        self.assertEqual(coll.find().count(), len(data) - 3)
        self.assertEqual(
            coll.find().filter({"name": {"$regex": "J.*"}}).remove(), 1
        )
        self.assertEqual(coll.find().count(), len(data) - 4)
        self.conn.commit()

    def test_3406(self):
        "3406 - test create and drop Index"
        index_name = "TestIndexes_ix_1"
        index_spec = {
            "name": index_name,
            "fields": [
                {"path": "address.city", "datatype": "string", "order": "asc"}
            ],
        }
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestIndexes")
        self.conn.commit()
        coll.dropIndex(index_name)
        coll.createIndex(index_spec)
        self.assertRaises(TypeError, coll.createIndex, 3)
        with self.assertRaisesFullCode("ORA-40733"):
            coll.createIndex(index_spec)
        self.assertTrue(coll.dropIndex(index_name))
        self.assertFalse(coll.dropIndex(index_name))

    def test_3407(self):
        "3407 - test getting documents from Collection"
        self.conn.autocommit = True
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestGetDocs")
        data = [
            {"name": "John", "address": {"city": "Bangalore"}},
            {"name": "Johnson", "address": {"city": "Banaras"}},
            {"name": "Joseph", "address": {"city": "Mangalore"}},
            {"name": "Jibin", "address": {"city": "Secunderabad"}},
            {"name": "Andrew", "address": {"city": "Hyderabad"}},
        ]
        inserted_keys = list(sorted(coll.insertOneAndGet(v).key for v in data))
        fetched_keys = list(
            sorted(doc.key for doc in coll.find().getDocuments())
        )
        self.assertEqual(fetched_keys, inserted_keys)

    def test_3408(self):
        "3408 - test fetching documents from a cursor"
        self.conn.autocommit = True
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestFindViaCursor")
        data = [
            {"name": "John", "address": {"city": "Bangalore"}},
            {"name": "Johnson", "address": {"city": "Banaras"}},
            {"name": "Joseph", "address": {"city": "Mangalore"}},
        ]
        inserted_keys = list(sorted(coll.insertOneAndGet(v).key for v in data))
        fetched_keys = list(sorted(doc.key for doc in coll.find().getCursor()))
        self.assertEqual(fetched_keys, inserted_keys)

    def test_3409(self):
        "3409 - test removing multiple documents using multiple keys"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestRemoveMultipleDocs")
        data = [
            {"name": "John", "address": {"city": "Bangalore"}},
            {"name": "Johnson", "address": {"city": "Banaras"}},
            {"name": "Joseph", "address": {"city": "Mangalore"}},
            {"name": "Jibin", "address": {"city": "Secunderabad"}},
            {"name": "Andrew", "address": {"city": "Hyderabad"}},
            {"name": "Matthew", "address": {"city": "Mumbai"}},
        ]
        docs = [coll.insertOneAndGet(v) for v in data]
        keys = [docs[i].key for i in (1, 3, 5)]
        num_removed = coll.find().keys(keys).remove()
        self.assertEqual(num_removed, len(keys))
        self.assertEqual(coll.find().count(), len(data) - len(keys))
        self.conn.commit()

    def test_3410(self):
        "3410 - test using version to get documents and remove them"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestDocumentVersion")
        content = {"name": "John", "address": {"city": "Bangalore"}}
        inserted_doc = coll.insertOneAndGet(content)
        key = inserted_doc.key
        version = inserted_doc.version
        doc = coll.find().key(key).version(version).getOne().getContent()
        self.__normalize_docs([doc])
        self.assertEqual(doc, content)
        new_content = {"name": "James", "address": {"city": "Delhi"}}
        replaced_doc = coll.find().key(key).replaceOneAndGet(new_content)
        new_version = replaced_doc.version
        doc = coll.find().key(key).version(version).getOne()
        self.assertIsNone(doc)
        doc = coll.find().key(key).version(new_version).getOne().getContent()
        self.__normalize_docs([doc])
        self.assertEqual(doc, new_content)
        self.assertEqual(coll.find().key(key).version(version).remove(), 0)
        self.assertEqual(coll.find().key(key).version(new_version).remove(), 1)
        self.assertEqual(coll.find().count(), 0)
        self.conn.commit()

    def test_3411(self):
        "3411 - test keys with GetCursor"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestKeysWithGetCursor")
        values_to_insert = [
            {"name": "John", "address": {"city": "Bangalore"}},
            {"name": "Johnson", "address": {"city": "Banaras"}},
            {"name": "Joseph", "address": {"city": "Mangalore"}},
            {"name": "Jibin", "address": {"city": "Secunderabad"}},
            {"name": "Andrew", "address": {"city": "Hyderabad"}},
            {"name": "Matthew", "address": {"city": "Mumbai"}},
        ]
        docs = [coll.insertOneAndGet(value) for value in values_to_insert]
        keys = [docs[i].key for i in (2, 4, 5)]
        fetched_keys = [doc.key for doc in coll.find().keys(keys).getCursor()]
        self.assertEqual(list(sorted(fetched_keys)), list(sorted(keys)))
        self.conn.commit()

    def test_3412(self):
        "3412 - test createdOn attribute of Document"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("CreatedOn")
        data = {"name": "John", "address": {"city": "Bangalore"}}
        doc = coll.insertOneAndGet(data)
        self.assertEqual(doc.createdOn, doc.lastModified)

    @unittest.skipIf(
        test_env.get_client_version() < (20, 1), "unsupported client"
    )
    def test_3413(self):
        "3413 - test Soda truncate"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestTruncateDocs")
        values_to_insert = [
            {"name": "George", "age": 47},
            {"name": "Susan", "age": 39},
            {"name": "John", "age": 50},
            {"name": "Jill", "age": 54},
        ]
        for value in values_to_insert:
            coll.insertOne(value)
        self.conn.commit()
        self.assertEqual(coll.find().count(), len(values_to_insert))
        coll.truncate()
        self.assertEqual(coll.find().count(), 0)

    def test_3414(self):
        "3414 - verify hints are reflected in the executed SQL statement"
        soda_db = self.get_soda_database()
        cursor = self.conn.cursor()
        statement = """
                SELECT
                    ( SELECT t2.sql_fulltext
                      FROM v$sql t2
                      WHERE t2.sql_id = t1.prev_sql_id
                        AND t2.child_number = t1.prev_child_number
                    )
                FROM v$session t1
                WHERE t1.audsid = sys_context('userenv', 'sessionid')"""
        coll = soda_db.createCollection("TestSodaHint")
        values_to_insert = [
            {"name": "George", "age": 47},
            {"name": "Susan", "age": 39},
        ]
        coll.insertOneAndGet(values_to_insert[0], hint="MONITOR")
        cursor.execute(statement)
        (result,) = cursor.fetchone()
        self.assertIn("MONITOR", result.read())

        coll.find().hint("MONITOR").getOne().getContent()
        cursor.execute(statement)
        (result,) = cursor.fetchone()
        self.assertIn("MONITOR", result.read())

        coll.insertOneAndGet(values_to_insert[1], hint="NO_MONITOR")
        cursor.execute(statement)
        (result,) = cursor.fetchone()
        self.assertIn("NO_MONITOR", result.read())

    def test_3415(self):
        "3415 - test error for invalid type for soda hint"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("InvalidSodaHint")
        self.assertRaises(
            TypeError, coll.insertOneAndGet, dict(name="Fred", age=16), hint=5
        )
        self.assertRaises(
            TypeError,
            coll.insertManyAndGet,
            dict(name="George", age=25),
            hint=10,
        )
        self.assertRaises(
            TypeError, coll.saveAndGet, dict(name="Sally", age=36), hint=5
        )
        self.assertRaises(TypeError, coll.find().hint, 2)

    def test_3416(self):
        "3416 - test name and metadata attribute"
        soda_db = self.get_soda_database()
        collection_name = "TestCollectionMetadata"
        coll = soda_db.createCollection(collection_name)
        self.assertEqual(coll.name, collection_name)
        self.assertEqual(coll.metadata["tableName"], collection_name)

    def test_3417(self):
        "3417 - test insertMany"
        soda_db = self.get_soda_database(minclient=(18, 5))
        coll = soda_db.createCollection("TestInsertMany")
        values_to_insert = [
            dict(name="George", age=25),
            soda_db.createDocument(dict(name="Lucas", age=47)),
        ]
        coll.insertMany(values_to_insert)
        self.conn.commit()
        fetched_values = [doc.getContent() for doc in coll.find().getCursor()]
        fetched_values.sort(key=lambda x: x["name"])
        for fetched_val, expected_val in zip(fetched_values, values_to_insert):
            if not isinstance(expected_val, dict):
                expected_val = expected_val.getContent()
            self.assertEqual(fetched_val, fetched_val)
        with self.assertRaisesFullCode("DPI-1031"):
            coll.insertMany([])

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "save() is not implemented in Oracle Database 23ai",
    )
    def test_3418(self):
        "3418 - test save"
        soda_db = self.get_soda_database(minclient=(19, 9))
        coll = soda_db.createCollection("TestSodaSave")
        values_to_save = [
            dict(name="Jill", age=37),
            soda_db.createDocument(dict(name="John", age=7)),
            soda_db.createDocument(dict(name="Charles", age=24)),
        ]
        for value in values_to_save:
            coll.save(value)
        self.conn.commit()
        fetched_docs = coll.find().getDocuments()
        for fetched_doc, expected_doc in zip(fetched_docs, values_to_save):
            if isinstance(expected_doc, dict):
                expected_doc = soda_db.createDocument(expected_doc)
            self.assertEqual(
                fetched_doc.getContent(), expected_doc.getContent()
            )

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "save() is not implemented in Oracle Database 23ai",
    )
    def test_3419(self):
        "3419 - test saveAndGet with hint"
        soda_db = self.get_soda_database(minclient=(19, 11))
        cursor = self.conn.cursor()
        statement = """
                SELECT
                    ( SELECT t2.sql_fulltext
                      FROM v$sql t2
                      WHERE t2.sql_id = t1.prev_sql_id
                        AND t2.child_number = t1.prev_child_number
                    )
                FROM v$session t1
                WHERE t1.audsid = sys_context('userenv', 'sessionid')"""
        coll = soda_db.createCollection("TestSodaSaveWithHint")

        values_to_save = [
            dict(name="Jordan", age=59),
            dict(name="Curry", age=34),
        ]
        hints = ["MONITOR", "NO_MONITOR"]
        for value, hint in zip(values_to_save, hints):
            coll.saveAndGet(value, hint=hint)
            coll.find().hint(hint).getOne().getContent()
            cursor.execute(statement)
            (result,) = cursor.fetchone()
            self.assertIn(hint, result.read())

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "save() is not implemented in Oracle Database 23ai",
    )
    def test_3420(self):
        "3420 - test saveAndGet"
        soda_db = self.get_soda_database(minclient=(19, 9))
        coll = soda_db.createCollection("TestSodaSaveAndGet")
        values_to_save = [
            dict(name="John", age=50),
            soda_db.createDocument(dict(name="Mark", age=45)),
            soda_db.createDocument(dict(name="Jill", age=32)),
        ]
        inserted_keys = []
        for value in values_to_save:
            doc = coll.saveAndGet(value)
            inserted_keys.append(doc.key)
        fetched_docs = coll.find().getDocuments()
        self.conn.commit()
        self.assertEqual(coll.find().count(), len(values_to_save))
        for key, fetched_doc in zip(inserted_keys, fetched_docs):
            doc = coll.find().key(key).getOne()
            self.assertEqual(doc.getContent(), fetched_doc.getContent())

    def test_3421(self):
        "3421 - test insert many and get"
        soda_db = self.get_soda_database(minclient=(18, 5))
        for name in soda_db.getCollectionNames():
            soda_db.openCollection(name).drop()
        coll = soda_db.createCollection("TestInsertManyAndGet")
        values_to_insert = [
            dict(name="George", age=25),
            soda_db.createDocument(dict(name="Lucas", age=47)),
        ]
        docs = coll.insertManyAndGet(values_to_insert)
        inserted_keys = [doc.key for doc in docs]
        self.conn.commit()
        self.assertEqual(coll.find().count(), len(values_to_insert))
        for key, expected_doc in zip(inserted_keys, values_to_insert):
            if isinstance(expected_doc, dict):
                expected_doc = soda_db.createDocument(expected_doc)
            doc = coll.find().key(key).getOne().getContent()
            self.__normalize_docs([doc])
            self.assertEqual(doc, expected_doc.getContent())

    def test_3422(self):
        "3422 - close document cursor and confirm exception is raised"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestCloseSodaDocCursor")
        cursor = coll.find().getCursor()
        cursor.close()
        with self.assertRaisesFullCode("DPY-1006"):
            cursor.close()
        with self.assertRaisesFullCode("DPY-1006"):
            next(cursor)

    def test_3423(self):
        "3423 - test limit to get specific amount of documents"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaLimit")
        values_to_insert = [{"group": "Camila"} for i in range(20)]
        coll.insertMany(values_to_insert)
        self.conn.commit()
        docs = coll.find().getDocuments()
        self.assertEqual(len(docs), len(values_to_insert))
        docs = coll.find().limit(3).getDocuments()
        self.assertEqual(len(docs), 3)

    def test_3424(self):
        "3424 - get count exceptions when using limit and skip"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaCountExceptions")
        data = [{"song": "WYMCA"} for i in range(20)]
        coll.insertMany(data)
        self.conn.commit()
        with self.assertRaisesFullCode("ORA-40748"):
            coll.find().limit(5).count()
        with self.assertRaisesFullCode("ORA-40748"):
            coll.find().skip(10).count()

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "map mode not supported with native collections in Oracle Database 23",
    )
    def test_3425(self):
        "3425 - test mapMode parameter"
        soda_db = self.get_soda_database()
        data = [{"price": 4900}, {"price": 8}]
        expected_data = data * 2

        original_coll = soda_db.createCollection("TestCollMapMode")
        original_coll.insertMany(data)
        mapped_coll = soda_db.createCollection("TestCollMapMode", mapMode=True)
        mapped_coll.insertMany(data)

        for coll in [original_coll, mapped_coll]:
            fetched_data = list(
                doc.getContent() for doc in coll.find().getDocuments()
            )
            self.__normalize_docs(fetched_data)
            self.assertEqual(fetched_data, expected_data)
            with self.assertRaisesFullCode("ORA-40626"):
                coll.drop()
        self.conn.commit()
        self.assertTrue(original_coll.drop())
        self.assertFalse(mapped_coll.drop())

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "map mode not supported with native collections in Oracle Database 23",
    )
    def test_3426(self):
        "3426 - test mapping a new collection from an non-existent table"
        soda_db = self.get_soda_database()
        with self.assertRaisesFullCode("ORA-40623"):
            soda_db.createCollection("TestSodaMapNonExistent", mapMode=True)

    def test_3427(self):
        "3427 - test negative cases for SodaOperation methods"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaOperationNegative")
        self.assertRaises(TypeError, coll.find().filter, 5)
        self.assertRaises(TypeError, coll.find().key, 2)
        self.assertRaises(TypeError, coll.find().keys, [1, 2, 3])
        self.assertRaises(TypeError, coll.find().skip, "word")
        self.assertRaises(TypeError, coll.find().skip, -5)
        self.assertRaises(TypeError, coll.find().version, 1971)
        self.assertRaises(TypeError, coll.find().limit, "a word")

    def test_3428(self):
        "3428 - test fetchArraySize"
        soda_db = self.get_soda_database(minclient=(19, 5))
        coll = soda_db.createCollection("TestSodaFetchArraySize")
        for i in range(90):
            coll.insertOne({"name": "Emmanuel", "age": i + 1})
        self.conn.commit()

        self.setup_round_trip_checker()
        # setting array size to 0 will use the default value of 100
        # requires a single round-trip
        coll.find().fetchArraySize(0).getDocuments()
        self.assertRoundTrips(1)

        # setting array size to 1 requires a round-trip for each SodaDoc
        coll.find().fetchArraySize(1).getDocuments()
        self.assertRoundTrips(91)

        # setting array size to 20 requires 5 round-trips
        coll.find().fetchArraySize(20).getDocuments()
        self.assertRoundTrips(5)

        # getting a SodaDocCursor requires a round-trip
        coll.find().fetchArraySize(0).getCursor()
        self.assertRoundTrips(1)

        # setting array size to 1 and iterating the SodaDocCursor requires a
        # round-trip for each SodaDoc
        soda_doc_cursor = coll.find().fetchArraySize(1).getCursor()
        for soda_doc in soda_doc_cursor:
            continue
        self.assertRoundTrips(91)

        # setting array size to 50 and iterating the SodaDocCursor requires
        # two round-trips
        soda_doc_cursor = coll.find().fetchArraySize(50).getCursor()
        for soda_doc in soda_doc_cursor:
            continue
        self.assertRoundTrips(2)

        # check a few negative scenarios
        self.assertRaises(TypeError, coll.find().fetchArraySize, "Mijares")
        self.assertRaises(TypeError, coll.find().fetchArraySize, -1)

    def test_3429(self):
        "3429 - test getting indexes on a collection"
        soda_db = self.get_soda_database(minclient=(19, 13))
        coll = soda_db.createCollection("TestSodaListIndexes")
        index_1 = {
            "name": "ix_3428-1",
            "fields": [
                {"path": "address.city", "datatype": "string", "order": "asc"}
            ],
        }
        index_2 = {
            "name": "ix_3428-2",
            "fields": [
                {
                    "path": "address.postal_code",
                    "datatype": "string",
                    "order": "asc",
                }
            ],
        }
        self.assertEqual(coll.listIndexes(), [])
        coll.createIndex(index_1)
        coll.createIndex(index_2)
        indexes = coll.listIndexes()
        indexes.sort(key=lambda x: x["name"])
        self.assertEqual(indexes[0]["fields"][0]["path"], "address.city")
        self.assertEqual(
            indexes[1]["fields"][0]["path"], "address.postal_code"
        )

    def test_3430(self):
        "3430 - test locking documents on fetch"
        soda_db = self.get_soda_database(minclient=(19, 11))
        coll = soda_db.createCollection("TestSodaLockDocs")
        values_to_insert = [
            {"name": "Bob", "age": 46},
            {"name": "Barb", "age": 45},
            {"name": "Sandy", "age": 47},
        ]
        coll.insertMany(values_to_insert)
        coll.find().lock().getDocuments()

    def test_3431(self):
        "3431 - test that drop returns the correct boolean"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestDropCollection")
        self.assertTrue(coll.drop())

        # the collection has already been dropped
        self.assertFalse(coll.drop())

    @unittest.skipIf(
        test_env.get_client_version() > (23, 0),
        "map mode not supported with native collections in Oracle Database 23",
    )
    def test_3432(self):
        "3432 - test drop with an empty mapped collection"
        soda_db = self.get_soda_database()
        original_coll = soda_db.createCollection("TestDropMapMode")
        mapped_coll = soda_db.createCollection("TestDropMapMode", mapMode=True)
        self.assertTrue(mapped_coll.drop())
        self.assertFalse(original_coll.drop())

    def test_3433(self):
        "3433 - test that replaceOne() returns a correct boolean"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestReplaceDocReturns")
        doc = coll.insertOneAndGet({"address": {"city": "Sydney"}})

        new_content = {"address": {"city": "Melbourne"}}
        self.assertTrue(coll.find().key(doc.key).replaceOne(new_content))

        unregistered_key = "DB4A2628F1E0985C891F3F4836"
        self.assertFalse(
            coll.find().key(unregistered_key).replaceOne(new_content)
        )
        self.conn.commit()

    def test_3434(self):
        "3434 - replaceOne() and replaceOneAndGet() with invalid scenarios"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestReplaceOneNegative")
        coll.insertMany([{"Wisdom": 1.7} for d in range(2)])
        keys = [d.key for d in coll.find().getDocuments()]
        with self.assertRaisesFullCode("ORA-40734"):
            coll.find().keys(keys).replaceOne({"data": "new"})
        with self.assertRaisesFullCode("ORA-40734"):
            coll.find().keys(keys).replaceOneAndGet({"data": "new"})

    def test_3435(self):
        "3435 - test writting a read-only collection"
        soda_db = self.get_soda_database()
        metadata = {
            "readOnly": True,
        }
        coll = soda_db.createCollection("TestCollReadOnly", metadata)

        methods = [
            coll.insertOne,
            coll.insertOneAndGet,
            coll.insertMany,
            coll.insertManyAndGet,
            coll.save,
            coll.saveAndGet,
        ]
        for method in methods:
            with self.subTest(method=method):
                with self.assertRaisesFullCode("ORA-40663"):
                    method({"Song 1": "No end"})

    def test_3436(self):
        "3436 - createCollection() with the same name and metadata"
        soda_db = self.get_soda_database()
        coll_name = "TestCollSameMetadata"
        coll1 = soda_db.createCollection(coll_name, {"readOnly": True})
        coll2 = soda_db.createCollection(coll_name, {"readOnly": True})
        self.assertTrue(coll1.drop())
        self.assertFalse(coll2.drop())

    def test_3437(self):
        "3437 - createCollection() with the same name but different metadata"
        soda_db = self.get_soda_database()
        coll_name = "TestCollDifferentMetadata"
        coll = soda_db.createCollection(coll_name)
        with self.assertRaisesFullCode("ORA-40669"):
            soda_db.createCollection(coll_name, {"readOnly": False})
        coll.drop()

        coll = soda_db.createCollection(coll_name, {"readOnly": True})
        with self.assertRaisesFullCode("ORA-40669"):
            soda_db.createCollection(coll_name, {"readOnly": False})

    def test_3438(self):
        "3438 - test getDataGuide() with an index with data-guide support"
        self.conn = test_env.get_connection()
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaDataGuideEnabled")
        data = [
            {
                "team": "backend",
                "created_in": 2001,
                "members": [{"developer": "Joseph"}, {"tester": "Mark"}],
            },
            {"team": "frontend", "area": "user interface"},
        ]
        coll.insertMany(data)
        self.conn.commit()
        index = {
            "name": "ix_3438",
            "dataguide": "on",
        }
        coll.createIndex(index)

        data_guide = coll.getDataGuide().getContent()

        client_version = test_env.get_client_version()
        server_version = test_env.get_server_version()
        if client_version >= (23, 4) and server_version >= (23, 4):
            self.assertEqual(data_guide["properties"]["_id"]["type"], "id")

        values = [
            ("team", "string"),
            ("created_in", "number"),
            ("area", "string"),
        ]
        for name, typ in values:
            self.assertEqual(data_guide["properties"][name]["type"], typ)
            self.assertRegex(
                data_guide["properties"][name]["o:preferred_column_name"],
                f"(JSON_DOCUMENT|DATA)\\${name}",
            )
        self.assertEqual(data_guide["properties"]["members"]["type"], "array")

        members_values = [
            ("tester", "string", 4),
            ("developer", "string", 8),
        ]
        for name, typ, length in members_values:
            members_items = data_guide["properties"]["members"]["items"]
            self.assertEqual(members_items["properties"][name]["type"], typ)
            self.assertEqual(
                members_items["properties"][name]["o:length"], length
            )
            self.assertRegex(
                members_items["properties"][name]["o:preferred_column_name"],
                f"(JSON_DOCUMENT|DATA)\\${name}",
            )

    def test_3439(self):
        "3439 - test getDataGuide() with an index without data-guide support"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaDataGuideDisabled")

        coll.insertOne({"data": "test_3439"})
        self.conn.commit()
        index = {
            "name": "ix-3439",
            "dataguide": "off",
        }
        coll.createIndex(index)
        with self.assertRaisesFullCode("ORA-40582"):
            coll.getDataGuide()

    def test_3440(self):
        "3440 - test getDataGuide() with an empty collection"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestDataGuideWithEmptyColl")
        coll.createIndex({"name": "ix_3440", "dataguide": "on"})
        self.assertIsNone(coll.getDataGuide())

    def test_3441(self):
        "3441 - test getDataGuide() without a json search index"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestSodaDataGuideWithoutIndex")
        with self.assertRaisesFullCode("ORA-40582"):
            coll.getDataGuide()

    def test_3442(self):
        "3442 - test mapMode parameter with metadata"
        soda_db = self.get_soda_database()
        data = [{"price": 4900}, {"price": 8}]
        expected_data = data * 2
        coll_name = "TestCollOriginal"

        metadata = {
            "tableName": coll_name,
            "keyColumn": {"name": "ID"},
            "contentColumn": {"name": "JSON_DOCUMENT", "sqlType": "BLOB"},
            "versionColumn": {"name": "VERSION", "method": "UUID"},
            "lastModifiedColumn": {"name": "LAST_MODIFIED"},
            "creationTimeColumn": {"name": "CREATED_ON"},
        }

        original_coll = soda_db.createCollection(coll_name, metadata=metadata)
        original_coll.insertMany(data)
        mapped_coll = soda_db.createCollection(
            "TestCollMapMode", metadata=metadata, mapMode=True
        )
        mapped_coll.insertMany(data)

        for coll in [original_coll, mapped_coll]:
            fetched_data = list(
                doc.getContent() for doc in coll.find().getDocuments()
            )
            self.__normalize_docs(fetched_data)
            self.assertEqual(fetched_data, expected_data)

        with self.assertRaisesFullCode("ORA-40626"):
            original_coll.drop()
        self.assertTrue(mapped_coll.drop())
        self.conn.commit()
        self.assertTrue(original_coll.drop())
        self.assertFalse(mapped_coll.drop())

    def test_3443(self):
        "3443 - test mapping a new collection from an non-existent table"
        soda_db = self.get_soda_database()
        metadata = {"tableName": "TestNone"}
        with self.assertRaisesFullCode("ORA-40623"):
            soda_db.createCollection(
                "TestSodaMapNonExistent", metadata=metadata, mapMode=True
            )

    def test_3444(self):
        "3444 - test collections with mixture of media types"
        soda_db = self.get_soda_database()
        metadata = dict(mediaTypeColumn=dict(name="media_type"))
        coll = soda_db.createCollection("TestMixedMedia", metadata=metadata)
        test_data = [
            (dict(name="George", age=28), "application/json"),
            ("Sample Text", "text/plain"),
            (b"\x57\x25\xfe\x34\x56", "application/octet-stream"),
        ]
        for value, media_type in test_data:
            coll.find().remove()
            coll.insertOne(soda_db.createDocument(value, mediaType=media_type))
            fetched_doc = coll.find().getDocuments()[0]
            self.assertEqual(fetched_doc.mediaType, media_type)
            if media_type == "application/json":
                self.assertEqual(fetched_doc.getContent(), value)
                self.assertEqual(
                    json.loads(fetched_doc.getContentAsString()), value
                )
                self.assertEqual(
                    json.loads(fetched_doc.getContentAsBytes().decode()), value
                )
            elif media_type == "text/plain":
                self.assertEqual(fetched_doc.getContent(), value.encode())
                self.assertEqual(fetched_doc.getContentAsString(), value)
                self.assertEqual(
                    fetched_doc.getContentAsBytes(), value.encode()
                )
            else:
                self.assertEqual(fetched_doc.getContent(), value)
                self.assertEqual(fetched_doc.getContentAsBytes(), value)
                self.assertRaises(
                    UnicodeDecodeError, fetched_doc.getContentAsString
                )

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        or test_env.get_server_version() < (23, 4),
        "unsupported data types",
    )
    def test_3445(self):
        "3445 - test fetching documents with JSON data using extended types"
        soda_db = self.get_soda_database()
        val = {
            "testKey1": "testValue1",
            "testKey2": decimal.Decimal("12.78"),
            "testKey3": datetime.datetime(2023, 7, 3, 11, 10, 24),
        }
        doc = soda_db.createDocument(val)
        self.assertEqual(doc.getContent(), val)
        coll = soda_db.createCollection("TestJSONExtendedTypes")
        coll.insertOne(doc)
        fetched_doc = coll.find().getDocuments()[0]
        fetched_content = fetched_doc.getContent()
        self.__normalize_docs([fetched_content])
        self.assertEqual(fetched_content, val)
        self.assertEqual(doc.getContent(), val)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        or test_env.get_server_version() < (23, 4),
        "unsupported data types",
    )
    def test_3446(self):
        "3446 - test round-trip of JsonId"
        soda_db = self.get_soda_database()
        coll = soda_db.createCollection("TestJsonId")
        val = {
            "key1": 5,
            "key2": "A string",
            "key3": b"Raw data",
            "key4": datetime.datetime(2024, 3, 2, 10, 1, 36),
        }
        doc = soda_db.createDocument(val)
        coll.insertOne(doc)
        self.conn.commit()
        fetched_doc = coll.find().getDocuments()[0]
        fetched_content = fetched_doc.getContent()
        self.assertIs(type(fetched_content["_id"]), oracledb.JsonId)
        updated_val = val.copy()
        updated_val["key1"] = 25
        content = fetched_content.copy()
        content["key1"] = updated_val["key1"]
        updated_doc = soda_db.createDocument(content)
        coll.find().key(fetched_doc.key).replaceOne(updated_doc)
        fetched_doc = coll.find().getDocuments()[0]
        self.assertEqual(fetched_doc.getContent(), content)

    def test_3447(self):
        "3447 - test getting documents with client-assigned keys"
        soda_db = self.conn.getSodaDatabase()
        metadata = {"keyColumn": {"assignmentMethod": "client"}}
        coll = soda_db.createCollection(
            "TestSearchByClientAssignedKeys", metadata
        )
        test_values = [
            ("doc1", {"name": "Help others", "files": []}),
            ("doc2", {"name": "Family", "files": ["kids.txt"]}),
            ("doc3", {"name": "Our pets", "files": ["dogs.pdf"]}),
        ]
        docs = [soda_db.createDocument(d, k) for k, d in test_values]
        coll.insertMany(docs)

        for key, data in test_values:
            (fetched_doc,) = coll.find().key(key).getDocuments()
            self.assertEqual(fetched_doc.getContent(), data)

        keys = [key for key, _ in test_values]
        fetched_docs = coll.find().keys(keys).getDocuments()
        self.assertEqual(len(fetched_docs), 3)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# constants.pxi
#
# Cython file defining constants used by the base implementation classes
# (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

# base JSON constants
cdef enum:
    TNS_JSON_MAGIC_BYTE_1 = 0xff
    TNS_JSON_MAGIC_BYTE_2 = 0x4a        # 'J'
    TNS_JSON_MAGIC_BYTE_3 = 0x5a        # 'Z'
    TNS_JSON_VERSION_MAX_FNAME_255 = 1
    TNS_JSON_VERSION_MAX_FNAME_65535 = 3
    TNS_JSON_FLAG_HASH_ID_UINT8 = 0x0100
    TNS_JSON_FLAG_NUM_FNAMES_UINT16 = 0x0400
    TNS_JSON_FLAG_FNAMES_SEG_UINT32 = 0x0800
    TNS_JSON_FLAG_TINY_NODES_STAT = 0x2000
    TNS_JSON_FLAG_TREE_SEG_UINT32 = 0x1000
    TNS_JSON_FLAG_REL_OFFSET_MODE = 0x01
    TNS_JSON_FLAG_INLINE_LEAF = 0x02
    TNS_JSON_FLAG_LEN_IN_PCODE = 0x04
    TNS_JSON_FLAG_NUM_FNAMES_UINT32 = 0x08
    TNS_JSON_FLAG_IS_SCALAR = 0x10
    TNS_JSON_FLAG_SEC_FNAMES_SEG_UINT16 = 0x0100

# JSON data types
cdef enum:
    TNS_JSON_TYPE_NULL = 0x30
    TNS_JSON_TYPE_TRUE = 0x31
    TNS_JSON_TYPE_FALSE = 0x32
    TNS_JSON_TYPE_STRING_LENGTH_UINT8 = 0x33
    TNS_JSON_TYPE_NUMBER_LENGTH_UINT8 = 0x34
    TNS_JSON_TYPE_BINARY_DOUBLE = 0x36
    TNS_JSON_TYPE_STRING_LENGTH_UINT16 = 0x37
    TNS_JSON_TYPE_STRING_LENGTH_UINT32 = 0x38
    TNS_JSON_TYPE_TIMESTAMP = 0x39
    TNS_JSON_TYPE_BINARY_LENGTH_UINT16 = 0x3a
    TNS_JSON_TYPE_BINARY_LENGTH_UINT32 = 0x3b
    TNS_JSON_TYPE_DATE = 0x3c
    TNS_JSON_TYPE_INTERVAL_YM = 0x3d
    TNS_JSON_TYPE_INTERVAL_DS = 0x3e
    TNS_JSON_TYPE_TIMESTAMP_TZ = 0x7c
    TNS_JSON_TYPE_TIMESTAMP7 = 0x7d
    TNS_JSON_TYPE_ID = 0x7e
    TNS_JSON_TYPE_BINARY_FLOAT = 0x7f
    TNS_JSON_TYPE_OBJECT = 0x84
    TNS_JSON_TYPE_ARRAY = 0xc0
    TNS_JSON_TYPE_EXTENDED = 0x7b
    TNS_JSON_TYPE_VECTOR = 0x01

# VECTOR constants
cdef enum:
    TNS_VECTOR_MAGIC_BYTE = 0xDB
    TNS_VECTOR_VERSION_BASE = 0
    TNS_VECTOR_VERSION_WITH_BINARY = 1

# VECTOR flags
cdef enum:
    TNS_VECTOR_FLAG_NORM = 0x0002
    TNS_VECTOR_FLAG_NORM_RESERVED = 0x0010

# general constants
cdef enum:
    TNS_MAX_SHORT_LENGTH = 252
    TNS_DURATION_MID = 0x80000000
    TNS_DURATION_OFFSET = 60
    TNS_CHUNK_SIZE = 32767
    TNS_HAS_REGION_ID = 0x80

# timezone offsets
cdef enum:
    TZ_HOUR_OFFSET = 20
    TZ_MINUTE_OFFSET = 60



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# statement.pyx
#
# Cython file defining the Statement and BindInfo classes used to hold
# information about statements that are executed and any bind parameters that
# may be bound to those statements (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class BindInfo:

    cdef:
        uint32_t num_elements
        bint _is_return_bind
        uint8_t ora_type_num
        uint32_t buffer_size
        int16_t precision
        uint8_t bind_dir
        uint32_t size
        str _bind_name
        bint is_array
        int16_t scale
        uint8_t csfrm
        ThinVarImpl _bind_var_impl

    def __cinit__(self, str name, bint is_return_bind):
        self._bind_name = name
        self._is_return_bind = is_return_bind

    cdef BindInfo copy(self):
        return BindInfo(self._bind_name, self._is_return_bind)


cdef class StatementParser(BaseParser):

    cdef:
        bint returning_keyword_found

    cdef int _parse_bind_name(self, Statement stmt) except -1:
        """
        Bind variables are identified as follows:
        - Quoted and non-quoted bind names are allowed.
        - Quoted bind names can contain any characters.
        - Non-quoted bind names must begin with an alphabetic character.
        - Non-quoted bind names can only contain alphanumeric characters, the
          underscore, the dollar sign and the pound sign.
        - Non-quoted bind names cannot be Oracle Database Reserved Names (this
          is left to the server to detct and return an appropriate error).
        """
        cdef:
            bint quoted_name = False, in_bind = False, digits_only = False
            ssize_t start_pos = 0
            str bind_name
            Py_UCS4 ch
        self.temp_pos = self.pos + 1
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not in_bind:
                if cpython.Py_UNICODE_ISSPACE(ch):
                    self.temp_pos += 1
                    continue
                elif ch == '"':
                    quoted_name = True
                elif cpython.Py_UNICODE_ISDIGIT(ch):
                    digits_only = True
                elif not cpython.Py_UNICODE_ISALPHA(ch):
                    break
                in_bind = True
                start_pos = self.temp_pos
            elif digits_only and not cpython.Py_UNICODE_ISDIGIT(ch):
                self.pos = self.temp_pos - 1
                break
            elif quoted_name and ch == '"':
                self.pos = self.temp_pos
                break
            elif not digits_only and not quoted_name \
                    and not cpython.Py_UNICODE_ISALNUM(ch) \
                    and ch not in ('_', '$', '#'):
                self.pos = self.temp_pos - 1
                break
            self.temp_pos += 1
        if in_bind:
            if quoted_name:
                bind_name = stmt._sql[start_pos + 1:self.temp_pos]
            elif digits_only:
                bind_name = stmt._sql[start_pos:self.temp_pos]
            else:
                bind_name = stmt._sql[start_pos:self.temp_pos].upper()
            stmt._add_bind(bind_name)

    cdef int _parse_multiple_line_comment(self) except -1:
        """
        Multiple line comments consist of the characters /* followed by all
        characters up until */. This method is called when the first slash is
        detected and checks for the subsequent asterisk. If found, the comment
        is traversed and the current position is updated; otherwise, the
        current position is left untouched.
        """
        cdef:
            bint in_comment = False, exiting_comment = False
            Py_UCS4 ch
        self.temp_pos = self.pos + 1
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not in_comment:
                if ch != '*':
                    break
                in_comment = True
            elif ch == '*':
                exiting_comment = True
            elif exiting_comment:
                if ch == '/':
                    self.pos = self.temp_pos
                    break
                exiting_comment = False
            self.temp_pos += 1

    cdef int _parse_qstring(self) except -1:
        """
        Parses a q-string which consists of the characters "q" and a single
        quote followed by a start separator, any text that does not contain the
        end seprator and the end separator and ending quote. The following are
        examples that demonstrate this:
            - q'[...]'
            - q'{...}'
            - q'<...>'
            - q'(...)'
            - q'?...?' (where ? is any character)
        """
        cdef:
            bint exiting_qstring = False, in_qstring = False
            Py_UCS4 ch, sep = 0
        self.temp_pos += 1
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not in_qstring:
                if ch == '[':
                    sep = ']'
                elif ch == '{':
                    sep = '}'
                elif ch == '<':
                    sep = '>'
                elif ch == '(':
                    sep = ')'
                else:
                    sep = ch
                in_qstring = True
            elif not exiting_qstring and ch == sep:
                exiting_qstring = True
            elif exiting_qstring:
                if ch == "'":
                    self.pos = self.temp_pos
                    return 0
                elif ch != sep:
                    exiting_qstring = False
            self.temp_pos += 1
        errors._raise_err(errors.ERR_MISSING_QUOTE_IN_STRING)

    cdef int _parse_single_line_comment(self) except -1:
        """
        Single line comments consist of two dashes and all characters up to the
        next line break (or the end of the data). This method is called when
        the first dash is detected and checks for the subsequent dash. If
        found, the single line comment is traversed and the current position is
        updated; otherwise, the current position is left untouched.
        """
        cdef:
            bint in_comment = False
            Py_UCS4 ch
        self.temp_pos = self.pos + 1
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not in_comment:
                if ch != '-':
                    return 0
                in_comment = True
            elif cpython.Py_UNICODE_ISLINEBREAK(ch):
                break
            self.temp_pos += 1
        self.pos = self.temp_pos

    cdef int parse(self, Statement stmt) except -1:
        """
        Parses the SQL stored in the statement in order to determine the
        keyword that identifies the type of SQL being executed as well as a
        list of bind variable names. A check is also made for DML returning
        statements since the bind variables following the "INTO" keyword are
        treated differently from other bind variables.
        """
        cdef:
            bint initial_keyword_found = False, last_was_string = False
            Py_UCS4 ch, last_ch = 0, alpha_start_ch = 0
            ssize_t alpha_start_pos = 0, alpha_len
            bint last_was_alpha = False, is_alpha
            str keyword

        # initialization
        self.initialize(stmt._sql)

        # scan all characters in the string
        while self.pos < self.num_chars:
            self.temp_pos = self.pos
            ch = self.get_current_char()

            # look for certain keywords (initial keyword and the ones for
            # detecting DML returning statements
            is_alpha = cpython.Py_UNICODE_ISALPHA(ch)
            if is_alpha and not last_was_alpha:
                alpha_start_pos = self.pos
                alpha_start_ch = ch
            elif not is_alpha and last_was_alpha:
                alpha_len = self.pos - alpha_start_pos
                if not initial_keyword_found:
                    keyword = stmt._sql[alpha_start_pos:self.pos].upper()
                    stmt._determine_statement_type(keyword)
                    if stmt._is_ddl:
                        break
                    initial_keyword_found = True
                elif stmt._is_dml and not self.returning_keyword_found \
                        and alpha_len == 9 and alpha_start_ch in ('r', 'R'):
                    keyword = stmt._sql[alpha_start_pos:self.pos].upper()
                    if keyword == "RETURNING":
                        self.returning_keyword_found = True
                elif self.returning_keyword_found and alpha_len == 4 \
                        and alpha_start_ch in ('i', 'I'):
                    keyword = stmt._sql[alpha_start_pos:self.pos].upper()
                    if keyword == "INTO":
                        stmt._is_returning = True

            # need to keep track of whether the last token parsed was a string
            # (excluding whitespace) as if the last token parsed was a string
            # a following colon is not a bind variable but a part of the JSON
            # constant syntax
            if ch == "'":
                last_was_string = True
                if last_ch in ('q', 'Q'):
                    self._parse_qstring()
                else:
                    self.temp_pos += 1
                    self.parse_quoted_string(ch)
                    self.pos -= 1
            elif not cpython.Py_UNICODE_ISSPACE(ch):
                if ch == '-':
                    self._parse_single_line_comment()
                elif ch == '/':
                    self._parse_multiple_line_comment()
                elif ch == '"':
                    self.temp_pos += 1
                    self.parse_quoted_string(ch)
                    self.pos -= 1
                elif ch == ':' and not last_was_string:
                    self._parse_bind_name(stmt)
                last_was_string = False

            # advance to next character and track previous character
            self.pos += 1
            last_was_alpha = is_alpha
            last_ch = ch


cdef class Statement:

    cdef:
        str _sql
        bytes _sql_bytes
        uint32_t _sql_length
        uint16_t _cursor_id
        bint _is_query
        bint _is_plsql
        bint _is_dml
        bint _is_ddl
        bint _is_returning
        list _bind_info_list
        list _fetch_info_impls
        list _fetch_vars
        list _fetch_var_impls
        object _bind_info_dict
        object _last_output_type_handler
        uint32_t _num_columns
        bint _executed
        bint _binds_changed
        bint _no_prefetch
        bint _requires_define
        bint _return_to_cache
        bint _in_use

    cdef Statement copy(self):
        cdef:
            Statement copied_statement = Statement.__new__(Statement)
            object bind_info_dict
            BindInfo bind_info
        copied_statement._sql = self._sql
        copied_statement._sql_bytes = self._sql_bytes
        copied_statement._sql_length = self._sql_length
        copied_statement._is_query = self._is_query
        copied_statement._is_plsql = self._is_plsql
        copied_statement._is_dml = self._is_dml
        copied_statement._is_ddl = self._is_ddl
        copied_statement._is_returning = self._is_returning
        copied_statement._bind_info_list = \
                [bind_info.copy() for bind_info in self._bind_info_list]
        copied_statement._bind_info_dict = collections.OrderedDict()
        bind_info_dict = copied_statement._bind_info_dict
        for bind_info in copied_statement._bind_info_list:
            if bind_info._bind_name in bind_info_dict:
                bind_info_dict[bind_info._bind_name].append(bind_info)
            else:
                bind_info_dict[bind_info._bind_name] = [bind_info]
        copied_statement._return_to_cache = False
        return copied_statement

    cdef int _add_bind(self, str name) except -1:
        """
        Add bind information to the statement by examining the passed SQL for
        bind variable names.
        """
        cdef BindInfo info, orig_info
        if not self._is_plsql or name not in self._bind_info_dict:
            info = BindInfo(name, self._is_returning)
            self._bind_info_list.append(info)
            if info._bind_name in self._bind_info_dict:
                if self._is_returning:
                    orig_info = self._bind_info_dict[info._bind_name][-1]
                    if not orig_info._is_return_bind:
                        errors._raise_err(errors.ERR_DML_RETURNING_DUP_BINDS,
                                          name=info._bind_name)
                self._bind_info_dict[info._bind_name].append(info)
            else:
                self._bind_info_dict[info._bind_name] = [info]

    cdef _determine_statement_type(self, str sql_keyword):
        """
        Determine the type of the SQL statement by examining the first keyword
        found in the statement.
        """
        if sql_keyword in ("DECLARE", "BEGIN", "CALL"):
            self._is_plsql = True
        elif sql_keyword in ("SELECT", "WITH"):
            self._is_query = True
        elif sql_keyword in ("INSERT", "UPDATE", "DELETE", "MERGE"):
            self._is_dml = True
        elif sql_keyword in ("CREATE", "ALTER", "DROP", "GRANT", "REVOKE",
                             "ANALYZE", "AUDIT", "COMMENT", "TRUNCATE"):
            self._is_ddl = True

    cdef int _prepare(self, str sql) except -1:
        """
        Prepare the SQL for execution by determining the list of bind names
        that are found within it. The length of the SQL text is also calculated
        at this time.
        """
        cdef StatementParser parser = StatementParser.__new__(StatementParser)

        # retain normalized SQL (as string and bytes) as well as the length
        self._sql = sql
        self._sql_bytes = self._sql.encode()
        self._sql_length = <uint32_t> len(self._sql_bytes)

        # parse SQL and populate bind variable list (bind by position) and dict
        # (bind by name)
        self._bind_info_dict = collections.OrderedDict()
        self._bind_info_list = []
        parser.parse(self)

    cdef int _set_var(self, BindInfo bind_info, ThinVarImpl var_impl,
                      ThinCursorImpl cursor_impl) except -1:
        """
        Set the variable on the bind information and copy across metadata that
        will be used for binding. If the bind metadata has changed, mark the
        statement as requiring a full execute. In addition, binding a REF
        cursor also requires a full execute.
        """
        cdef object value
        if var_impl.dbtype._ora_type_num == TNS_DATA_TYPE_CURSOR:
            for value in var_impl._values:
                if value is not None and value._impl is cursor_impl:
                    errors._raise_err(errors.ERR_SELF_BIND_NOT_SUPPORTED)
            self._binds_changed = True
        if var_impl.dbtype._ora_type_num != bind_info.ora_type_num \
                or var_impl.size != bind_info.size \
                or var_impl.buffer_size != bind_info.buffer_size \
                or var_impl.precision != bind_info.precision \
                or var_impl.scale != bind_info.scale \
                or var_impl.is_array != bind_info.is_array \
                or var_impl.num_elements != bind_info.num_elements \
                or var_impl.dbtype._csfrm != bind_info.csfrm:
            bind_info.ora_type_num = var_impl.dbtype._ora_type_num
            bind_info.csfrm = var_impl.dbtype._csfrm
            bind_info.is_array = var_impl.is_array
            bind_info.num_elements = var_impl.num_elements
            bind_info.size = var_impl.size
            bind_info.buffer_size = var_impl.buffer_size
            bind_info.precision = var_impl.precision
            bind_info.scale = var_impl.scale
            self._binds_changed = True
        bind_info._bind_var_impl = var_impl

    cdef int clear_all_state(self) except -1:
        """
        Clears all state associated with the cursor.
        """
        self._fetch_vars = None
        self._fetch_info_impls = None
        self._fetch_var_impls = None
        self._executed = False
        self._binds_changed = False
        self._requires_define = False
        self._no_prefetch = False
        self._cursor_id = 0

    cdef bint requires_single_execute(self):
        """
        Returns a boolean indicating if the statement requires a single execute
        in order to be processed correctly by the server. If a PL/SQL block has
        not been executed before, the determination of input/output binds has
        not been completed and so a single execution is required in order to
        complete that determination.
        """
        return self._is_plsql and (self._cursor_id == 0 or self._binds_changed)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
1400 - Module for testing date/time variables
"""

import datetime

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            base_date = datetime.datetime(2002, 12, 9)
            date_interval = datetime.timedelta(
                days=i, hours=i * 2, minutes=i * 24
            )
            date_col = base_date + date_interval
            if i % 2:
                date_interval = datetime.timedelta(
                    days=i * 2, hours=i * 3, minutes=i * 36
                )
                nullable_col = base_date + date_interval
            else:
                nullable_col = None
            data_tuple = (i, date_col, nullable_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_1400(self):
        "1400 - test binding in a date"
        self.cursor.execute(
            "select * from TestDates where DateCol = :value",
            value=datetime.datetime(2002, 12, 13, 9, 36, 0),
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[4]])

    def test_1401(self):
        "1401 - test binding in a datetime.datetime value"
        self.cursor.execute(
            "select * from TestDates where DateCol = :value",
            value=datetime.datetime(2002, 12, 13, 9, 36, 0),
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[4]])

    def test_1402(self):
        "1402 - test binding date in a datetime variable"
        var = self.cursor.var(oracledb.DATETIME)
        date_val = datetime.date.today()
        var.setvalue(0, date_val)
        self.cursor.execute("select :1 from dual", [var])
        (result,) = self.cursor.fetchone()
        self.assertEqual(result.date(), date_val)

    def test_1403(self):
        "1403 - test binding in a date after setting input sizes to a string"
        self.cursor.setinputsizes(value=15)
        self.cursor.execute(
            "select * from TestDates where DateCol = :value",
            value=datetime.datetime(2002, 12, 14, 12, 0, 0),
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_1404(self):
        "1404 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute(
            "select * from TestDates where DateCol = :value",
            value=None,
        )
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1405(self):
        "1405 - test binding in a date array"
        array = [r[1] for r in self.raw_data]
        return_value = self.cursor.callfunc(
            "pkg_TestDateArrays.TestInArrays",
            oracledb.DB_TYPE_NUMBER,
            [5, datetime.date(2002, 12, 12), array],
        )
        self.assertEqual(return_value, 35.5)
        array += array[:5]
        return_value = self.cursor.callfunc(
            "pkg_TestDateArrays.TestInArrays",
            oracledb.DB_TYPE_NUMBER,
            [7, datetime.date(2002, 12, 13), array],
        )
        self.assertEqual(return_value, 24.0)

    def test_1406(self):
        "1406 - test binding in a date array (with setinputsizes)"
        return_value = self.cursor.var(oracledb.NUMBER)
        self.cursor.setinputsizes(array=[oracledb.DATETIME, 10])
        array = [r[1] for r in self.raw_data]
        self.cursor.execute(
            """
            begin
                :return_value := pkg_TestDateArrays.TestInArrays(
                    :start_value, :base_date, :array);
            end;
            """,
            return_value=return_value,
            start_value=6,
            base_date=oracledb.Date(2002, 12, 13),
            array=array,
        )
        self.assertEqual(return_value.getvalue(), 26.5)

    def test_1407(self):
        "1407 - test binding in a date array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.DATETIME, 10, 20)
        array.setvalue(0, [r[1] for r in self.raw_data])
        self.cursor.execute(
            """
            begin
                :return_value := pkg_TestDateArrays.TestInArrays(
                    :start_value, :base_date, :array);
            end;
            """,
            return_value=return_value,
            start_value=7,
            base_date=oracledb.Date(2002, 12, 14),
            array=array,
        )
        self.assertEqual(return_value.getvalue(), 17.5)

    def test_1408(self):
        "1408 - test binding in/out a date array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DATETIME, 10, 100)
        original_data = [r[1] for r in self.raw_data]
        array.setvalue(0, original_data)
        self.cursor.execute(
            """
            begin
                pkg_TestDateArrays.TestInOutArrays(:num_elems, :array);
            end;
            """,
            num_elems=5,
            array=array,
        )
        expected_value = [
            datetime.datetime(2002, 12, 17, 2, 24, 0),
            datetime.datetime(2002, 12, 18, 4, 48, 0),
            datetime.datetime(2002, 12, 19, 7, 12, 0),
            datetime.datetime(2002, 12, 20, 9, 36, 0),
            datetime.datetime(2002, 12, 21, 12, 0, 0),
        ] + original_data[5:]
        self.assertEqual(array.getvalue(), expected_value)

    def test_1409(self):
        "1409 - test binding out a date array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DATETIME, 6, 100)
        self.cursor.execute(
            """
            begin
                pkg_TestDateArrays.TestOutArrays(:num_elems, :array);
            end;
            """,
            num_elems=6,
            array=array,
        )
        expected_value = [
            datetime.datetime(2002, 12, 13, 4, 48, 0),
            datetime.datetime(2002, 12, 14, 9, 36, 0),
            datetime.datetime(2002, 12, 15, 14, 24, 0),
            datetime.datetime(2002, 12, 16, 19, 12, 0),
            datetime.datetime(2002, 12, 18, 0, 0, 0),
            datetime.datetime(2002, 12, 19, 4, 48, 0),
        ]
        self.assertEqual(array.getvalue(), expected_value)

    def test_1410(self):
        "1410 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute(
            """
            begin
                :value := to_date(20021209, 'YYYYMMDD');
            end;
            """
        )
        self.assertEqual(
            bind_vars["value"].getvalue(), datetime.datetime(2002, 12, 9)
        )

    def test_1411(self):
        "1411 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute(
            """
            begin
                :value := :value + 5.25;
            end;
            """,
            value=datetime.datetime(2002, 12, 12, 10, 0, 0),
        )
        self.assertEqual(
            bind_vars["value"].getvalue(),
            datetime.datetime(2002, 12, 17, 16, 0, 0),
        )

    def test_1412(self):
        "1412 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DATETIME)
        self.cursor.execute(
            """
            begin
                :value := to_date('20021231 12:31:00', 'YYYYMMDD HH24:MI:SS');
            end;
            """,
            value=var,
        )
        self.assertEqual(
            var.getvalue(), datetime.datetime(2002, 12, 31, 12, 31, 0)
        )

    def test_1413(self):
        "1413 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DATETIME)
        var.setvalue(0, datetime.datetime(2002, 12, 9, 6, 0, 0))
        self.cursor.execute(
            """
            begin
                :value := :value + 5.25;
            end;
            """,
            value=var,
        )
        self.assertEqual(
            var.getvalue(), datetime.datetime(2002, 12, 14, 12, 0, 0)
        )

    def test_1414(self):
        "1414 - test cursor description is accurate"
        self.cursor.execute("select * from TestDates")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ("DATECOL", oracledb.DB_TYPE_DATE, 23, None, None, None, False),
            ("NULLABLECOL", oracledb.DB_TYPE_DATE, 23, None, None, None, True),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1415(self):
        "1415 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestDates order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1416(self):
        "1416 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestDates order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_1417(self):
        "1417 - test that fetching a single row returns the correct results"
        self.cursor.execute(
            """
            select *
            from TestDates
            where IntCol in (3, 4)
            order by IntCol
            """
        )
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertIsNone(self.cursor.fetchone())

    def test_1418(self):
        "1418 - test fetching a date with year < 0"
        with self.assertRaises(ValueError):
            self.cursor.execute(
                "select to_date('-4712-01-01', 'SYYYY-MM-DD') from dual"
            )
            self.cursor.fetchone()


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# vector.pyx
#
# Cython file defining the classes and methods used for encoding and decoding
# VECTOR data (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef array.array float_template = array.array('f')
cdef array.array double_template = array.array('d')
cdef array.array int8_template = array.array('b')
cdef array.array uint8_template = array.array('B')

@cython.final
cdef class VectorDecoder(Buffer):

    cdef object decode(self, bytes data):
        """
        Returns a Python object corresponding to the encoded VECTOR bytes.
        """
        cdef:
            uint8_t magic_byte, version, vector_format, element_size = 0
            uint8_t * uint8_buf = NULL
            double *double_buf = NULL
            int8_t *int8_buf = NULL
            uint32_t num_elements, i
            float *float_buf = NULL
            const char_type* ptr
            array.array result
            uint16_t flags
            object value

        # populate the buffer with the data
        self._populate_from_bytes(data)

        # parse header
        self.read_ub1(&magic_byte)
        if magic_byte != TNS_VECTOR_MAGIC_BYTE:
            errors._raise_err(errors.ERR_UNEXPECTED_DATA,
                              data=bytes([magic_byte]))
        self.read_ub1(&version)
        if version > TNS_VECTOR_VERSION_WITH_BINARY:
            errors._raise_err(errors.ERR_VECTOR_VERSION_NOT_SUPPORTED,
                              version=version)
        self.read_uint16(&flags)
        self.read_ub1(&vector_format)
        self.read_uint32(&num_elements)
        if vector_format == VECTOR_FORMAT_FLOAT32:
            result = array.clone(float_template, num_elements, False)
            float_buf = result.data.as_floats
            element_size = 4
        elif vector_format == VECTOR_FORMAT_FLOAT64:
            result = array.clone(double_template, num_elements, False)
            double_buf = result.data.as_doubles
            element_size = 8
        elif vector_format == VECTOR_FORMAT_INT8:
            result = array.clone(int8_template, num_elements, False)
            int8_buf = result.data.as_schars
            element_size = 1
        elif vector_format == VECTOR_FORMAT_BINARY:
            num_elements = num_elements // 8
            result = array.clone(uint8_template, num_elements, False)
            uint8_buf = result.data.as_uchars
        else:
            errors._raise_err(errors.ERR_VECTOR_FORMAT_NOT_SUPPORTED,
                              vector_format=vector_format)
        if flags & TNS_VECTOR_FLAG_NORM_RESERVED \
                or flags & TNS_VECTOR_FLAG_NORM:
            self.skip_raw_bytes(8)

        # parse data
        for i in range(num_elements):
            if vector_format == VECTOR_FORMAT_FLOAT32:
                ptr = self._get_raw(element_size)
                self.parse_binary_float(ptr, &float_buf[i])
            elif vector_format == VECTOR_FORMAT_FLOAT64:
                ptr = self._get_raw(element_size)
                self.parse_binary_double(ptr, &double_buf[i])
            elif vector_format == VECTOR_FORMAT_INT8:
                self.read_sb1(&int8_buf[i])
            else:
                self.read_ub1(&uint8_buf[i])
        return result


@cython.final
cdef class VectorEncoder(GrowableBuffer):

    cdef int encode(self, array.array value) except -1:
        """
        Encodes the given value to the internal VECTOR format.
        """
        cdef:
            uint16_t flags = TNS_VECTOR_FLAG_NORM_RESERVED
            uint8_t vector_format, vector_version
            double *double_ptr = NULL
            uint8_t *uint8_ptr = NULL
            uint32_t num_elements, i
            float *float_ptr = NULL
            int8_t *int8_ptr = NULL
            object element

        # determine the type of vector to write
        if value.typecode == 'd':
            vector_format = VECTOR_FORMAT_FLOAT64
            double_ptr = value.data.as_doubles
        elif value.typecode == 'f':
            vector_format = VECTOR_FORMAT_FLOAT32
            float_ptr = value.data.as_floats
        elif value.typecode == 'b':
            vector_format = VECTOR_FORMAT_INT8
            int8_ptr = value.data.as_schars
        else:
            vector_format = VECTOR_FORMAT_BINARY
            uint8_ptr = value.data.as_uchars

        # determine vector version and number of elements
        if vector_format == VECTOR_FORMAT_BINARY:
            num_elements = (<uint32_t> len(value)) * 8
            vector_version = TNS_VECTOR_VERSION_WITH_BINARY
        else:
            num_elements = <uint32_t> len(value)
            vector_version = TNS_VECTOR_VERSION_BASE
            flags |= TNS_VECTOR_FLAG_NORM

        # write header
        self.write_uint8(TNS_VECTOR_MAGIC_BYTE)
        self.write_uint8(vector_version)
        self.write_uint16(flags)
        self.write_uint8(vector_format)
        self.write_uint32(num_elements)
        self._reserve_space(8)              # reserve space for norm

        # write elements
        if vector_format == VECTOR_FORMAT_INT8:
            self.write_raw(<char_type*> int8_ptr, num_elements)
        elif vector_format == VECTOR_FORMAT_BINARY:
            self.write_raw(<char_type*> uint8_ptr, num_elements // 8)
        else:
            for i in range(num_elements):
                if vector_format == VECTOR_FORMAT_FLOAT32:
                    self.write_binary_float(float_ptr[i], write_length=False)
                elif vector_format == VECTOR_FORMAT_FLOAT64:
                    self.write_binary_double(double_ptr[i], write_length=False)
